# Copyright (C) 2005-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import warnings

from breezy import bugtracker, revision
from breezy.revision import NULL_REVISION
from breezy.tests import TestCase, TestCaseWithTransport
from breezy.tests.matchers import MatchesAncestry

# We're allowed to test deprecated interfaces
warnings.filterwarnings(
    "ignore",
    ".*get_intervening_revisions was deprecated",
    DeprecationWarning,
    r"breezy\.tests\.test_revision",
)

# XXX: Make this a method of a merge base case


def make_branches(self, format=None):
    """Create two branches.

    branch 1 has 6 commits, branch 2 has 3 commits
    commit 10 is a ghosted merge merge from branch 1

    the object graph is
    B:     A:
    a..0   a..0
    a..1   a..1
    a..2   a..2
    b..3   a..3 merges b..4
    b..4   a..4
    b..5   a..5 merges b..5
    b..6 merges a4

    so A is missing b6 at the start
    and B is missing a3, a4, a5
    """
    tree1 = self.make_branch_and_tree("branch1", format=format)
    br1 = tree1.branch

    tree1.commit("Commit one", rev_id=b"a@u-0-0")
    tree1.commit("Commit two", rev_id=b"a@u-0-1")
    tree1.commit("Commit three", rev_id=b"a@u-0-2")

    tree2 = tree1.controldir.sprout("branch2").open_workingtree()
    br2 = tree2.branch
    tree2.commit("Commit four", rev_id=b"b@u-0-3")
    tree2.commit("Commit five", rev_id=b"b@u-0-4")
    self.assertEqual(br2.last_revision(), b"b@u-0-4")

    tree1.merge_from_branch(br2)
    tree1.commit("Commit six", rev_id=b"a@u-0-3")
    tree1.commit("Commit seven", rev_id=b"a@u-0-4")
    tree2.commit("Commit eight", rev_id=b"b@u-0-5")
    self.assertEqual(br2.last_revision(), b"b@u-0-5")

    tree1.merge_from_branch(br2)
    tree1.commit("Commit nine", rev_id=b"a@u-0-5")
    # DO NOT MERGE HERE - we WANT a GHOST.
    br1.lock_read()
    try:
        graph = br1.repository.get_graph()
        revhistory = list(
            graph.iter_lefthand_ancestry(br1.last_revision(), [revision.NULL_REVISION])
        )
        revhistory.reverse()
    finally:
        br1.unlock()
    tree2.add_parent_tree_id(revhistory[4])
    tree2.commit("Commit ten - ghost merge", rev_id=b"b@u-0-6")

    return br1, br2


class TestIsAncestor(TestCaseWithTransport):
    def test_recorded_ancestry(self):
        """Test that commit records all ancestors."""
        br1, br2 = make_branches(self)
        d = [
            (b"a@u-0-0", [b"a@u-0-0"]),
            (b"a@u-0-1", [b"a@u-0-0", b"a@u-0-1"]),
            (b"a@u-0-2", [b"a@u-0-0", b"a@u-0-1", b"a@u-0-2"]),
            (b"b@u-0-3", [b"a@u-0-0", b"a@u-0-1", b"a@u-0-2", b"b@u-0-3"]),
            (b"b@u-0-4", [b"a@u-0-0", b"a@u-0-1", b"a@u-0-2", b"b@u-0-3", b"b@u-0-4"]),
            (
                b"a@u-0-3",
                [
                    b"a@u-0-0",
                    b"a@u-0-1",
                    b"a@u-0-2",
                    b"b@u-0-3",
                    b"b@u-0-4",
                    b"a@u-0-3",
                ],
            ),
            (
                b"a@u-0-4",
                [
                    b"a@u-0-0",
                    b"a@u-0-1",
                    b"a@u-0-2",
                    b"b@u-0-3",
                    b"b@u-0-4",
                    b"a@u-0-3",
                    b"a@u-0-4",
                ],
            ),
            (
                b"b@u-0-5",
                [
                    b"a@u-0-0",
                    b"a@u-0-1",
                    b"a@u-0-2",
                    b"b@u-0-3",
                    b"b@u-0-4",
                    b"b@u-0-5",
                ],
            ),
            (
                b"a@u-0-5",
                [
                    b"a@u-0-0",
                    b"a@u-0-1",
                    b"a@u-0-2",
                    b"a@u-0-3",
                    b"a@u-0-4",
                    b"b@u-0-3",
                    b"b@u-0-4",
                    b"b@u-0-5",
                    b"a@u-0-5",
                ],
            ),
            (
                b"b@u-0-6",
                [
                    b"a@u-0-0",
                    b"a@u-0-1",
                    b"a@u-0-2",
                    b"a@u-0-4",
                    b"b@u-0-3",
                    b"b@u-0-4",
                    b"b@u-0-5",
                    b"b@u-0-6",
                ],
            ),
        ]
        br1_only = (b"a@u-0-3", b"a@u-0-4", b"a@u-0-5")
        br2_only = (b"b@u-0-6",)
        for branch in br1, br2:
            for rev_id, anc in d:
                if rev_id in br1_only and branch is not br1:
                    continue
                if rev_id in br2_only and branch is not br2:
                    continue
                self.assertThat(anc, MatchesAncestry(branch.repository, rev_id))


class TestIntermediateRevisions(TestCaseWithTransport):
    def setUp(self):
        TestCaseWithTransport.setUp(self)
        self.br1, self.br2 = make_branches(self)
        wt1 = self.br1.controldir.open_workingtree()
        wt2 = self.br2.controldir.open_workingtree()
        wt2.commit("Commit eleven", rev_id=b"b@u-0-7")
        wt2.commit("Commit twelve", rev_id=b"b@u-0-8")
        wt2.commit("Commit thirtteen", rev_id=b"b@u-0-9")

        wt1.merge_from_branch(self.br2)
        wt1.commit("Commit fourtten", rev_id=b"a@u-0-6")

        wt2.merge_from_branch(self.br1)
        wt2.commit("Commit fifteen", rev_id=b"b@u-0-10")


class MockRevisionSource:
    """A RevisionSource that takes a pregenerated graph.

    This is useful for testing revision graph algorithms where
    the actual branch existing is irrelevant.
    """

    def __init__(self, full_graph):
        self._full_graph = full_graph

    def get_revision_graph_with_ghosts(self, revision_ids):
        # This is mocked out to just return a constant graph.
        return self._full_graph


class TestCommonAncestor(TestCaseWithTransport):
    """Test checking whether a revision is an ancestor of another revision."""

    def test_get_history(self):
        # TODO: test ghosts on the left hand branch's impact
        # TODO: test ghosts on all parents, we should get some
        # indicator. i.e. NULL_REVISION
        # RBC 20060608
        tree = self.make_branch_and_tree(".")
        tree.commit("1", rev_id=b"1", allow_pointless=True)
        tree.commit("2", rev_id=b"2", allow_pointless=True)
        tree.commit("3", rev_id=b"3", allow_pointless=True)
        rev = tree.branch.repository.get_revision(b"1")
        history = rev.get_history(tree.branch.repository)
        self.assertEqual([None, b"1"], history)
        rev = tree.branch.repository.get_revision(b"2")
        history = rev.get_history(tree.branch.repository)
        self.assertEqual([None, b"1", b"2"], history)
        rev = tree.branch.repository.get_revision(b"3")
        history = rev.get_history(tree.branch.repository)
        self.assertEqual([None, b"1", b"2", b"3"], history)


class TestReservedId(TestCase):
    def test_is_reserved_id(self):
        self.assertEqual(True, revision.is_reserved_id(NULL_REVISION))
        self.assertEqual(True, revision.is_reserved_id(revision.CURRENT_REVISION))
        self.assertEqual(True, revision.is_reserved_id(b"arch:"))
        self.assertEqual(False, revision.is_reserved_id(b"null"))
        self.assertEqual(
            False, revision.is_reserved_id(b"arch:a@example.com/c--b--v--r")
        )
        self.assertEqual(False, revision.is_reserved_id(None))


class TestRevisionMethods(TestCase):
    def test_get_summary(self):
        r = revision.Revision("1")
        r.message = "a"
        self.assertEqual("a", r.get_summary())
        r.message = "a\nb"
        self.assertEqual("a", r.get_summary())
        r.message = "\na\nb"
        self.assertEqual("a", r.get_summary())
        r.message = None
        self.assertEqual("", r.get_summary())

    def test_get_apparent_authors(self):
        r = revision.Revision("1")
        r.committer = "A"
        self.assertEqual(["A"], r.get_apparent_authors())
        r.properties["author"] = "B"
        self.assertEqual(["B"], r.get_apparent_authors())
        r.properties["authors"] = "C\nD"
        self.assertEqual(["C", "D"], r.get_apparent_authors())

    def test_get_apparent_authors_no_committer(self):
        r = revision.Revision("1")
        self.assertEqual([], r.get_apparent_authors())


class TestRevisionBugs(TestCase):
    """Tests for getting the bugs that a revision is linked to."""

    def test_no_bugs(self):
        r = revision.Revision("1")
        self.assertEqual([], list(r.iter_bugs()))

    def test_some_bugs(self):
        r = revision.Revision(
            "1",
            properties={
                "bugs": bugtracker.encode_fixes_bug_urls(
                    [
                        ("http://example.com/bugs/1", "fixed"),
                        ("http://launchpad.net/bugs/1234", "fixed"),
                    ]
                )
            },
        )
        self.assertEqual(
            [
                ("http://example.com/bugs/1", bugtracker.FIXED),
                ("http://launchpad.net/bugs/1234", bugtracker.FIXED),
            ],
            list(r.iter_bugs()),
        )

    def test_no_status(self):
        r = revision.Revision("1", properties={"bugs": "http://example.com/bugs/1"})
        self.assertRaises(bugtracker.InvalidLineInBugsProperty, list, r.iter_bugs())

    def test_too_much_information(self):
        r = revision.Revision(
            "1", properties={"bugs": "http://example.com/bugs/1 fixed bar"}
        )
        self.assertRaises(bugtracker.InvalidLineInBugsProperty, list, r.iter_bugs())

    def test_invalid_status(self):
        r = revision.Revision(
            "1", properties={"bugs": "http://example.com/bugs/1 faxed"}
        )
        self.assertRaises(bugtracker.InvalidBugStatus, list, r.iter_bugs())



# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>
# Copyright (C) 2006-2009 Canonical Ltd

# Authors: Robert Collins <robert.collins@canonical.com>
#          Jelmer Vernooij <jelmer@jelmer.uk>
#          John Carr <john.carr@unrouted.co.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""A GIT branch and repository format implementation for bzr."""

import os
import sys

dulwich_minimum_version = (0, 19, 11)

from .. import __version__ as breezy_version
from .. import errors as brz_errors
from .. import trace, urlutils
from ..commands import plugin_cmds
from ..controldir import ControlDirFormat, Prober, format_registry
from ..controldir import network_format_registry as controldir_network_format_registry
from ..transport import (
    register_lazy_transport,
    register_transport_proto,
    transport_server_registry,
)

if getattr(sys, "frozen", None):
    # allow import additional libs from ./_lib for bzr.exe only
    sys.path.append(os.path.normpath(os.path.join(os.path.dirname(__file__), "_lib")))


def import_dulwich():
    try:
        from dulwich import __version__ as dulwich_version
    except ModuleNotFoundError:
        raise brz_errors.DependencyNotPresent(
            "dulwich", "bzr-git: Please install dulwich, https://www.dulwich.io/"
        )
    else:
        if dulwich_version < dulwich_minimum_version:
            raise brz_errors.DependencyNotPresent(
                "dulwich",
                "bzr-git: Dulwich is too old; at least "
                f"{dulwich_minimum_version[0]}.{dulwich_minimum_version[1]}.{dulwich_minimum_version[2]} is required",
            )


_versions_checked = False


def lazy_check_versions():
    global _versions_checked
    if _versions_checked:
        return
    import_dulwich()
    _versions_checked = True


format_registry.register_lazy(
    "git",
    __name__ + ".dir",
    "LocalGitControlDirFormat",
    help="GIT repository.",
    native=False,
    experimental=False,
)

format_registry.register_lazy(
    "git-bare",
    __name__ + ".dir",
    "BareLocalGitControlDirFormat",
    help="Bare GIT repository (no working tree).",
    native=False,
    experimental=False,
)

from ..revisionspec import RevisionSpec_dwim, revspec_registry

revspec_registry.register_lazy("git:", __name__ + ".revspec", "RevisionSpec_git")
RevisionSpec_dwim.append_possible_lazy_revspec(
    __name__ + ".revspec", "RevisionSpec_git"
)


class LocalGitProber(Prober):
    @classmethod
    def priority(klass, transport):
        return 10

    def probe_transport(self, transport):
        try:
            external_url = transport.external_url()
        except brz_errors.InProcessTransport:
            raise brz_errors.NotBranchError(path=transport.base)
        if external_url.startswith("http:") or external_url.startswith("https:"):
            # Already handled by RemoteGitProber
            raise brz_errors.NotBranchError(path=transport.base)
        if urlutils.split(transport.base)[1] == ".git":
            raise brz_errors.NotBranchError(path=transport.base)
        if not transport.has_any(["objects", ".git/objects", ".git"]):
            raise brz_errors.NotBranchError(path=transport.base)
        lazy_check_versions()
        from .dir import BareLocalGitControlDirFormat, LocalGitControlDirFormat

        if transport.has_any([".git/objects", ".git"]):
            return LocalGitControlDirFormat()
        if transport.has("info") and transport.has("objects"):
            return BareLocalGitControlDirFormat()
        raise brz_errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        from .dir import BareLocalGitControlDirFormat, LocalGitControlDirFormat

        return [BareLocalGitControlDirFormat(), LocalGitControlDirFormat()]


def user_agent_for_github():
    # GitHub requires we lie. https://github.com/dulwich/dulwich/issues/562
    return "git/Breezy/{}".format(breezy_version)


def is_github_url(url):
    (scheme, user, password, host, port, path) = urlutils.parse_url(url)
    return host in ("github.com", "gopkg.in")


class RemoteGitProber(Prober):
    @classmethod
    def priority(klass, transport):
        # This is a surprisingly good heuristic to determine whether this
        # prober is more likely to succeed than the Bazaar one.
        if "git" in transport.base:
            return -15
        return -10

    def probe_http_transport(self, transport):
        # This function intentionally doesn't use any of the support code under
        # breezy.git, since it's called for every repository that's
        # accessed over HTTP, whether it's Git, Bzr or something else.
        # Importing Dulwich and the other support code adds unnecessray slowdowns.
        base_url = urlutils.strip_segment_parameters(transport.external_url())
        url = urlutils.URL.from_string(base_url)
        url.user = url.quoted_user = None
        url.password = url.quoted_password = None
        url = urlutils.join(str(url), "info/refs") + "?service=git-upload-pack"
        headers = {
            "Content-Type": "application/x-git-upload-pack-request",
            "Accept": "application/x-git-upload-pack-result",
        }
        if is_github_url(url):
            # GitHub requires we lie.
            # https://github.com/dulwich/dulwich/issues/562
            headers["User-Agent"] = user_agent_for_github()
        resp = transport.request("GET", url, headers=headers)
        if resp.status in (404, 405):
            raise brz_errors.NotBranchError(transport.base)
        elif resp.status == 400 and resp.reason == "no such method: info":
            # hgweb :(
            raise brz_errors.NotBranchError(transport.base)
        elif resp.status != 200:
            raise brz_errors.UnexpectedHttpStatus(
                url, resp.status, headers=resp.getheaders()
            )

        ct = resp.getheader("Content-Type")
        if ct and ct.startswith("application/x-git"):
            from .remote import RemoteGitControlDirFormat

            return RemoteGitControlDirFormat()
        elif not ct:
            from .dir import BareLocalGitControlDirFormat

            ret = BareLocalGitControlDirFormat()
            ret._refs_text = resp.read()
            return ret
        raise brz_errors.NotBranchError(transport.base)

    def probe_transport(self, transport):
        try:
            external_url = transport.external_url()
        except brz_errors.InProcessTransport:
            raise brz_errors.NotBranchError(path=transport.base)

        if external_url.startswith("http:") or external_url.startswith("https:"):
            return self.probe_http_transport(transport)

        if not external_url.startswith("git://") and not external_url.startswith(
            "git+"
        ):
            raise brz_errors.NotBranchError(transport.base)

        # little ugly, but works
        from .remote import GitSmartTransport, RemoteGitControlDirFormat

        if isinstance(transport, GitSmartTransport):
            return RemoteGitControlDirFormat()
        raise brz_errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        from .remote import RemoteGitControlDirFormat

        return [RemoteGitControlDirFormat()]


ControlDirFormat.register_prober(LocalGitProber)
ControlDirFormat.register_prober(RemoteGitProber)

register_transport_proto("git://", help="Access using the Git smart server protocol.")
register_transport_proto(
    "git+ssh://", help="Access using the Git smart server protocol over SSH."
)

register_lazy_transport("git://", __name__ + ".remote", "TCPGitSmartTransport")
register_lazy_transport("git+ssh://", __name__ + ".remote", "SSHGitSmartTransport")


plugin_cmds.register_lazy("cmd_git_import", [], __name__ + ".commands")
plugin_cmds.register_lazy(
    "cmd_git_object", ["git-objects", "git-cat"], __name__ + ".commands"
)
plugin_cmds.register_lazy("cmd_git_refs", [], __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_apply", [], __name__ + ".commands")
plugin_cmds.register_lazy(
    "cmd_git_push_pristine_tar_deltas",
    ["git-push-pristine-tar", "git-push-pristine"],
    __name__ + ".commands",
)


def extract_git_foreign_revid(rev):
    try:
        foreign_revid = rev.foreign_revid
    except AttributeError:
        from .mapping import mapping_registry

        foreign_revid, mapping = mapping_registry.parse_revision_id(rev.revision_id)
        return foreign_revid
    else:
        from .mapping import foreign_vcs_git

        if rev.mapping.vcs == foreign_vcs_git:
            return foreign_revid
        else:
            raise brz_errors.InvalidRevisionId(rev.revision_id, None)


def update_stanza(rev, stanza):
    try:
        git_commit = extract_git_foreign_revid(rev)
    except brz_errors.InvalidRevisionId:
        pass
    else:
        stanza.add("git-commit", git_commit)


from ..hooks import install_lazy_named_hook

install_lazy_named_hook(
    "breezy.version_info_formats.format_rio",
    "RioVersionInfoBuilder.hooks",
    "revision",
    update_stanza,
    "git commits",
)


def rewrite_instead_of(location, purpose):
    from dulwich.config import StackedConfig, iter_instead_of

    config = StackedConfig.default()

    push = purpose != "read"

    longest_needle = ""
    updated_url = location
    for needle, replacement in iter_instead_of(config, push):
        if not location.startswith(needle):
            continue
        if len(longest_needle) < len(needle):
            longest_needle = needle
            if longest_needle == "lp:":
                # Leave the lp: prefix to the launchpad plugin, if loaded
                import breezy.plugins

                if hasattr(breezy.plugins, "launchpad"):
                    trace.warning(
                        "Ignoring insteadOf lp: in git config, because the Launchpad plugin is loaded."
                    )
                    continue
            updated_url = replacement + location[len(needle) :]
    return updated_url


from ..location import hooks as location_hooks

location_hooks.install_named_hook(
    "rewrite_location", rewrite_instead_of, "apply Git insteadOf / pushInsteadOf"
)

transport_server_registry.register_lazy(
    "git",
    __name__ + ".server",
    "serve_git",
    "Git Smart server protocol over TCP. (default port: 9418)",
)

transport_server_registry.register_lazy(
    "git-receive-pack",
    __name__ + ".server",
    "serve_git_receive_pack",
    help="Git Smart server receive pack command. (inetd mode only)",
)
transport_server_registry.register_lazy(
    "git-upload-pack",
    __name__ + "git.server",
    "serve_git_upload_pack",
    help="Git Smart server upload pack command. (inetd mode only)",
)

from ..repository import format_registry as repository_format_registry
from ..repository import network_format_registry as repository_network_format_registry

repository_network_format_registry.register_lazy(
    b"git", __name__ + ".repository", "GitRepositoryFormat"
)

register_extra_lazy_repository_format = repository_format_registry.register_extra_lazy
register_extra_lazy_repository_format(__name__ + ".repository", "GitRepositoryFormat")

from ..branch import network_format_registry as branch_network_format_registry

branch_network_format_registry.register_lazy(
    b"git", __name__ + ".branch", "LocalGitBranchFormat"
)


from ..branch import format_registry as branch_format_registry

branch_format_registry.register_extra_lazy(
    __name__ + ".branch",
    "LocalGitBranchFormat",
)
branch_format_registry.register_extra_lazy(
    __name__ + ".remote",
    "RemoteGitBranchFormat",
)


from ..workingtree import format_registry as workingtree_format_registry

workingtree_format_registry.register_extra_lazy(
    __name__ + ".workingtree",
    "GitWorkingTreeFormat",
)

controldir_network_format_registry.register_lazy(
    b"git", __name__ + ".dir", "GitControlDirFormat"
)


from ..diff import format_registry as diff_format_registry

diff_format_registry.register_lazy(
    "git", __name__ + ".send", "GitDiffTree", "Git am-style diff format"
)

from ..send import format_registry as send_format_registry

send_format_registry.register_lazy(
    "git", __name__ + ".send", "send_git", "Git am-style diff format"
)

from ..directory_service import directories

directories.register_lazy(
    "github:", __name__ + ".directory", "GitHubDirectory", "GitHub directory."
)
directories.register_lazy(
    "git@github.com:", __name__ + ".directory", "GitHubDirectory", "GitHub directory."
)

from ..help_topics import topic_registry

topic_registry.register_lazy(
    "git", __name__ + ".help", "help_git", "Using Bazaar with Git"
)

from ..foreign import foreign_vcs_registry

foreign_vcs_registry.register_lazy(
    "git", __name__ + ".mapping", "foreign_vcs_git", "Stupid content tracker"
)


def update_git_cache(repository, revid):
    """Update the git cache after a local commit."""
    if getattr(repository, "_git", None) is not None:
        return  # No need to update cache for git repositories

    if not repository.control_transport.has("git"):
        return  # No existing cache, don't bother updating
    try:
        lazy_check_versions()
    except brz_errors.DependencyNotPresent as e:
        # dulwich is probably missing. silently ignore
        trace.mutter("not updating git map for %r: %s", repository, e)

    from .object_store import BazaarObjectStore

    store = BazaarObjectStore(repository)
    with store.lock_write():
        try:
            parent_revisions = set(repository.get_parent_map([revid])[revid])
        except KeyError:
            # Isn't this a bit odd - how can a revision that was just committed
            # be missing?
            return
        missing_revisions = store._missing_revisions(parent_revisions)
        if not missing_revisions:
            store._cache.idmap.start_write_group()
            try:
                # Only update if the cache was up to date previously
                store._update_sha_map_revision(revid)
            except BaseException:
                store._cache.idmap.abort_write_group()
                raise
            else:
                store._cache.idmap.commit_write_group()


def post_commit_update_cache(
    local_branch, master_branch, old_revno, old_revid, new_revno, new_revid
):
    if local_branch is not None:
        update_git_cache(local_branch.repository, new_revid)
    update_git_cache(master_branch.repository, new_revid)


def loggerhead_git_hook(branch_app, environ):
    branch = branch_app.branch
    config_stack = branch.get_config_stack()
    if not config_stack.get("git.http"):
        return None
    from .server import git_http_hook

    return git_http_hook(branch, environ["REQUEST_METHOD"], environ["PATH_INFO"])


install_lazy_named_hook(
    "breezy.branch",
    "Branch.hooks",
    "post_commit",
    post_commit_update_cache,
    "git cache",
)
install_lazy_named_hook(
    "breezy.plugins.loggerhead.apps.branch",
    "BranchWSGIApp.hooks",
    "controller",
    loggerhead_git_hook,
    "git support",
)


from ..config import Option, bool_from_store, option_registry

option_registry.register(
    Option(
        "git.http",
        default=None,
        from_unicode=bool_from_store,
        invalid="warning",
        help="""\
Allow fetching of Git packs over HTTP.

This enables support for fetching Git packs over HTTP in Loggerhead.
""",
    )
)


def test_suite():
    from . import tests

    return tests.test_suite()



# Copyright (C) 2007-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Remote dirs, repositories and branches."""

import re

from dulwich.refs import SymrefLoop

from .. import config, debug, errors, osutils, trace, ui, urlutils
from ..controldir import BranchReferenceLoop
from ..errors import (
    AlreadyBranchError,
    BzrError,
    ConnectionReset,
    DivergedBranches,
    InProcessTransport,
    InvalidRevisionId,
    LockContention,
    NoSuchRevision,
    NoSuchTag,
    NotBranchError,
    NotLocalUrl,
    PermissionDenied,
    TransportError,
    UnexpectedHttpStatus,
    UninitializableFormat,
)
from ..push import PushResult
from ..revision import NULL_REVISION
from ..revisiontree import RevisionTree
from ..transport import NoSuchFile, Transport, register_urlparse_netloc_protocol
from . import is_github_url, lazy_check_versions, user_agent_for_github

lazy_check_versions()

import os
import select
import urllib.parse as urlparse

import dulwich
import dulwich.client
from dulwich.errors import GitProtocolError, HangupException
from dulwich.pack import (
    PACK_SPOOL_FILE_MAX_SIZE,
    Pack,
    load_pack_index,
    pack_objects_to_data,
)
from dulwich.refs import SYMREF, DictRefsContainer
from dulwich.repo import NotGitRepository

from .branch import (
    GitBranch,
    GitBranchFormat,
    GitBranchPushResult,
    GitTags,
    _quick_lookup_revno,
)
from .dir import GitControlDirFormat, GitDir
from .errors import GitSmartRemoteNotSupported
from .mapping import encode_git_path, mapping_registry
from .object_store import get_object_store
from .push import remote_divergence
from .refs import is_peeled, ref_to_tag_name, tag_name_to_ref
from .repository import GitRepository, GitRepositoryFormat

# urlparse only supports a limited number of schemes by default
register_urlparse_netloc_protocol("git")
register_urlparse_netloc_protocol("git+ssh")


class GitPushResult(PushResult):
    def _lookup_revno(self, revid):
        try:
            return _quick_lookup_revno(self.source_branch, self.target_branch, revid)
        except GitSmartRemoteNotSupported:
            return None

    @property
    def old_revno(self):
        return self._lookup_revno(self.old_revid)

    @property
    def new_revno(self):
        return self._lookup_revno(self.new_revid)


# Don't run any tests on GitSmartTransport as it is not intended to be
# a full implementation of Transport
def get_test_permutations():
    return []


def split_git_url(url):
    """Split a Git URL.

    :param url: Git URL
    :return: Tuple with host, port, username, path.
    """
    parsed_url = urlparse.urlparse(url)
    path = urlparse.unquote(parsed_url.path)
    if path.startswith("/~"):
        path = path[1:]
    return (parsed_url.hostname or "", parsed_url.port, parsed_url.username, path)


class RemoteGitError(BzrError):
    _fmt = "Remote server error: %(msg)s"


class ProtectedBranchHookDeclined(BzrError):
    _fmt = "Protected branch hook declined"


class HeadUpdateFailed(BzrError):
    _fmt = (
        "Unable to update remote HEAD branch. To update the master "
        "branch, specify the URL %(base_url)s,branch=master."
    )

    def __init__(self, base_url):
        super().__init__()
        self.base_url = base_url


def parse_git_error(url, message):
    """Parse a remote git server error and return a bzr exception.

    :param url: URL of the remote repository
    :param message: Message sent by the remote git server
    """
    message = str(message).strip()
    if (
        message.startswith("Could not find Repository ")
        or message == "Repository not found."
        or (message.startswith("Repository ") and message.endswith(" not found."))
    ):
        return NotBranchError(url, message)
    if message == "HEAD failed to update":
        base_url = urlutils.strip_segment_parameters(url)
        return HeadUpdateFailed(base_url)
    if message.startswith("access denied or repository not exported:"):
        extra, path = message.split(":", 1)
        return PermissionDenied(path.strip(), extra)
    if message.endswith("You are not allowed to push code to this project."):
        return PermissionDenied(url, message)
    if message.endswith(" does not appear to be a git repository"):
        return NotBranchError(url, message)
    if message == "A repository for this project does not exist yet.":
        return NotBranchError(url, message)
    if message == "pre-receive hook declined":
        return PermissionDenied(url, message)
    if re.match("(.+) is not a valid repository name", message.splitlines()[0]):
        return NotBranchError(url, message)
    if message == (
        "GitLab: You are not allowed to push code to protected branches "
        "on this project."
    ):
        return PermissionDenied(url, message)
    m = re.match(r"Permission to ([^ ]+) denied to ([^ ]+)\.", message)
    if m:
        return PermissionDenied(m.group(1), "denied to {}".format(m.group(2)))
    if message == "Host key verification failed.":
        return TransportError("Host key verification failed")
    if message == "[Errno 104] Connection reset by peer":
        return ConnectionReset(message)
    if message == "The remote server unexpectedly closed the connection.":
        return TransportError(message)
    m = re.match(r"unexpected http resp ([0-9]+) for (.*)", message)
    if m:
        # TODO(jelmer): Have dulwich raise an exception and look at that instead?
        return UnexpectedHttpStatus(
            path=m.group(2), code=int(m.group(1)), extra=message
        )
    if message == "protected branch hook declined":
        return ProtectedBranchHookDeclined(msg=message)
    # Don't know, just return it to the user as-is
    return RemoteGitError(message)


def parse_git_hangup(url, e):
    """Parse the error lines from a git servers stderr on hangup.

    :param url: URL of the remote repository
    :param e: A HangupException
    """
    stderr_lines = getattr(e, "stderr_lines", None)
    if not stderr_lines:
        return ConnectionReset("Connection closed early", e)
    if all(line.startswith(b"remote: ") for line in stderr_lines):
        stderr_lines = [line[len(b"remote: ") :] for line in stderr_lines]
    interesting_lines = [
        line for line in stderr_lines if line and line.replace(b"=", b"")
    ]
    if len(interesting_lines) == 1:
        interesting_line = interesting_lines[0]
        return parse_git_error(url, interesting_line.decode("utf-8", "surrogateescape"))
    return RemoteGitError(b"\n".join(stderr_lines).decode("utf-8", "surrogateescape"))


class GitSmartTransport(Transport):
    def __init__(self, url, _client=None):
        Transport.__init__(self, url)
        (self._host, self._port, self._username, self._path) = split_git_url(url)
        if "transport" in debug.debug_flags:
            trace.mutter(
                "host: %r, user: %r, port: %r, path: %r",
                self._host,
                self._username,
                self._port,
                self._path,
            )
        self._client = _client
        self._stripped_path = self._path.rsplit(",", 1)[0]

    def external_url(self):
        return self.base

    def has(self, relpath):
        return False

    def _get_client(self):
        raise NotImplementedError(self._get_client)

    def _get_path(self):
        return self._stripped_path

    def get(self, path):
        raise NoSuchFile(path)

    def abspath(self, relpath):
        return urlutils.join(self.base, relpath)

    def clone(self, offset=None):
        """See Transport.clone()."""
        if offset is None:
            newurl = self.base
        else:
            newurl = urlutils.join(self.base, offset)

        return self.__class__(newurl, self._client)


class TCPGitSmartTransport(GitSmartTransport):
    _scheme = "git"

    def _get_client(self):
        if self._client is not None:
            ret = self._client
            self._client = None
            return ret
        if self._host == "":
            # return dulwich.client.LocalGitClient()
            return dulwich.client.SubprocessGitClient()
        return dulwich.client.TCPGitClient(
            self._host, self._port, report_activity=self._report_activity
        )


class SSHSocketWrapper:
    def __init__(self, sock):
        self.sock = sock

    def read(self, len=None):
        return self.sock.recv(len)

    def write(self, data):
        return self.sock.write(data)

    def can_read(self):
        return len(select.select([self.sock.fileno()], [], [], 0)[0]) > 0


class DulwichSSHVendor(dulwich.client.SSHVendor):
    def __init__(self):
        from ..transport import ssh

        self.bzr_ssh_vendor = ssh._get_ssh_vendor()

    def run_command(self, host, command, username=None, port=None):
        connection = self.bzr_ssh_vendor.connect_ssh(
            username=username, password=None, port=port, host=host, command=command
        )
        (kind, io_object) = connection.get_sock_or_pipes()
        if kind == "socket":
            return SSHSocketWrapper(io_object)
        else:
            raise AssertionError("Unknown io object kind {!r}'".format(kind))


# dulwich.client.get_ssh_vendor = DulwichSSHVendor


class SSHGitSmartTransport(GitSmartTransport):
    _scheme = "git+ssh"

    def _get_path(self):
        path = self._stripped_path
        if path.startswith("/~/"):
            return path[3:]
        return path

    def _get_client(self):
        if self._client is not None:
            ret = self._client
            self._client = None
            return ret
        location_config = config.LocationConfig(self.base)
        client = dulwich.client.SSHGitClient(
            self._host,
            self._port,
            self._username,
            report_activity=self._report_activity,
        )
        # Set up alternate pack program paths
        upload_pack = location_config.get_user_option("git_upload_pack")
        if upload_pack:
            client.alternative_paths["upload-pack"] = upload_pack
        receive_pack = location_config.get_user_option("git_receive_pack")
        if receive_pack:
            client.alternative_paths["receive-pack"] = receive_pack
        return client


class RemoteGitBranchFormat(GitBranchFormat):
    def get_format_description(self):
        return "Remote Git Branch"

    @property
    def _matchingcontroldir(self):
        return RemoteGitControlDirFormat()

    def initialize(
        self, a_controldir, name=None, repository=None, append_revisions_only=None
    ):
        raise UninitializableFormat(self)


class DefaultProgressReporter:
    _GIT_PROGRESS_PARTIAL_RE = re.compile(r"(.*?): +(\d+)% \((\d+)/(\d+)\)")
    _GIT_PROGRESS_TOTAL_RE = re.compile(r"(.*?): (\d+)")

    def __init__(self, pb):
        self.pb = pb
        self.errors = []

    def progress(self, text):
        text = text.rstrip(b"\r\n")
        text = text.decode("utf-8", "surrogateescape")
        if text.lower().startswith("error: "):
            error = text[len("error: ") :]
            self.errors.append(error)
            trace.show_error("git: %s", error)
        else:
            trace.mutter("git: %s", text)
            g = self._GIT_PROGRESS_PARTIAL_RE.match(text)
            if g is not None:
                (text, pct, current, total) = g.groups()
                self.pb.update(text, int(current), int(total))
            else:
                g = self._GIT_PROGRESS_TOTAL_RE.match(text)
                if g is not None:
                    (text, total) = g.groups()
                    self.pb.update(text, None, int(total))
                else:
                    trace.note("%s", text)


_LOCK_REF_ERROR_MATCHER = re.compile("cannot lock ref '(.*)': (.*)")


class RemoteGitDir(GitDir):
    def __init__(self, transport, format, client, client_path):
        self._format = format
        self.root_transport = transport
        self.transport = transport
        self._mode_check_done = None
        self._client = client
        self._client_path = client_path
        self.base = self.root_transport.base
        self._refs = None

    @property
    def _gitrepository_class(self):
        return RemoteGitRepository

    def archive(
        self,
        format,
        committish,
        write_data,
        progress=None,
        write_error=None,
        subdirs=None,
        prefix=None,
        recurse_nested=False,
    ):
        if recurse_nested:
            raise NotImplementedError("recurse_nested is not yet supported")
        if progress is None:
            pb = ui.ui_factory.nested_progress_bar()
            progress = DefaultProgressReporter(pb).progress
        else:
            pb = None

        def progress_wrapper(message):
            if message.startswith(b"fatal: Unknown archive format '"):
                format = message.strip()[len(b"fatal: Unknown archive format '") : -1]
                raise errors.NoSuchExportFormat(format.decode("ascii"))
            return progress(message)

        try:
            self._client.archive(
                self._client_path,
                committish,
                write_data,
                progress_wrapper,
                write_error,
                format=(format.encode("ascii") if format else None),
                subdirs=subdirs,
                prefix=(encode_git_path(prefix) if prefix else None),
            )
        except HangupException as e:
            raise parse_git_hangup(self.transport.external_url(), e)
        except GitProtocolError as e:
            raise parse_git_error(self.transport.external_url(), e)
        finally:
            if pb is not None:
                pb.finished()

    def fetch_pack(self, determine_wants, graph_walker, pack_data, progress=None):
        if progress is None:
            pb = ui.ui_factory.nested_progress_bar()
            progress = DefaultProgressReporter(pb).progress
        else:
            pb = None
        try:
            result = self._client.fetch_pack(
                self._client_path, determine_wants, graph_walker, pack_data, progress
            )
            if result.refs is None:
                result.refs = {}
            self._refs = remote_refs_dict_to_container(result.refs, result.symrefs)
            return result
        except HangupException as e:
            raise parse_git_hangup(self.transport.external_url(), e)
        except GitProtocolError as e:
            raise parse_git_error(self.transport.external_url(), e)
        finally:
            if pb is not None:
                pb.finished()

    def send_pack(self, get_changed_refs, generate_pack_data, progress=None):
        if progress is None:
            pb = ui.ui_factory.nested_progress_bar()
            progress_reporter = DefaultProgressReporter(pb)
            progress = progress_reporter.progress
        else:
            progress_reporter = None
            pb = None

        def get_changed_refs_wrapper(remote_refs):
            if self._refs is not None:
                update_refs_container(self._refs, remote_refs)
            return get_changed_refs(remote_refs)

        try:
            result = self._client.send_pack(
                self._client_path,
                get_changed_refs_wrapper,
                generate_pack_data,
                progress,
            )
            for ref, msg in list(result.ref_status.items()):
                if msg:
                    result.ref_status[ref] = RemoteGitError(msg=msg)
            if progress_reporter:
                for error in progress_reporter.errors:
                    m = _LOCK_REF_ERROR_MATCHER.match(error)
                    if m:
                        result.ref_status[m.group(1)] = LockContention(
                            m.group(1), m.group(2)
                        )
            return result
        except HangupException as e:
            raise parse_git_hangup(self.transport.external_url(), e)
        except GitProtocolError as e:
            raise parse_git_error(self.transport.external_url(), e)
        finally:
            if pb is not None:
                pb.finished()

    def create_branch(
        self, name=None, repository=None, append_revisions_only=None, ref=None
    ):
        refname = self._get_selected_ref(name, ref)
        if refname != b"HEAD" and refname in self.get_refs_container():
            raise AlreadyBranchError(self.user_url)
        ref_chain, sha = self.get_refs_container().follow(self._get_selected_ref(name))
        if ref_chain and ref_chain[0] == b"HEAD" and len(ref_chain) > 1:
            refname = ref_chain[1]
        repo = self.open_repository()
        return RemoteGitBranch(self, repo, refname, sha)

    def destroy_branch(self, name=None):
        refname = self._get_selected_ref(name)

        def get_changed_refs(old_refs):
            ret = {}
            if refname not in old_refs:
                raise NotBranchError(self.user_url)
            ret[refname] = dulwich.client.ZERO_SHA
            return ret

        def generate_pack_data(have, want, ofs_delta=False, progress=None):
            return pack_objects_to_data([])

        result = self.send_pack(get_changed_refs, generate_pack_data)
        error = result.ref_status.get(refname)
        if error:
            raise error

    @property
    def user_url(self):
        return self.control_url

    @property
    def user_transport(self):
        return self.root_transport

    @property
    def control_url(self):
        return self.control_transport.base

    @property
    def control_transport(self):
        return self.root_transport

    def open_repository(self):
        return RemoteGitRepository(self)

    def get_branch_reference(self, name=None):
        ref = self._get_selected_ref(name)
        try:
            ref_chain, unused_sha = self.get_refs_container().follow(ref)
        except SymrefLoop:
            raise BranchReferenceLoop(self)
        if len(ref_chain) == 1:
            return None
        target_ref = ref_chain[1]
        from .refs import ref_to_branch_name

        try:
            branch_name = ref_to_branch_name(target_ref)
        except ValueError:
            params = {"ref": urlutils.quote(target_ref.decode("utf-8"), "")}
        else:
            if branch_name != "":
                params = {"branch": urlutils.quote(branch_name, "")}
            else:
                params = {}
        return urlutils.join_segment_parameters(self.user_url.rstrip("/"), params)

    def open_branch(
        self,
        name=None,
        unsupported=False,
        ignore_fallbacks=False,
        ref=None,
        possible_transports=None,
        nascent_ok=False,
    ):
        repo = self.open_repository()
        ref = self._get_selected_ref(name, ref)
        try:
            ref_chain, sha = self.get_refs_container().follow(ref)
        except SymrefLoop:
            raise BranchReferenceLoop(self)
        except NotGitRepository:
            raise NotBranchError(self.root_transport.base, controldir=self)
        if not nascent_ok and sha is None:
            raise NotBranchError(self.root_transport.base, controldir=self)
        return RemoteGitBranch(self, repo, ref_chain[-1], sha)

    def open_workingtree(self, recommend_upgrade=False):
        raise NotLocalUrl(self.transport.base)

    def has_workingtree(self):
        return False

    def get_peeled(self, name):
        return self.get_refs_container().get_peeled(name)

    def get_refs_container(self):
        if self._refs is not None:
            return self._refs
        result = self.fetch_pack(
            lambda x: None,
            None,
            lambda x: None,
            lambda x: trace.mutter("git: {}".format(x)),
        )
        self._refs = remote_refs_dict_to_container(result.refs, result.symrefs)
        return self._refs

    def push_branch(
        self,
        source,
        revision_id=None,
        overwrite=False,
        remember=False,
        create_prefix=False,
        lossy=False,
        name=None,
        tag_selector=None,
    ):
        """Push the source branch into this ControlDir."""
        if revision_id is None:
            # No revision supplied by the user, default to the branch
            # revision
            revision_id = source.last_revision()
        else:
            if not source.repository.has_revision(revision_id):
                raise NoSuchRevision(source, revision_id)

        push_result = GitPushResult()
        push_result.workingtree_updated = None
        push_result.master_branch = None
        push_result.source_branch = source
        push_result.stacked_on = None
        push_result.branch_push_result = None
        repo = self.find_repository()
        refname = self._get_selected_ref(name)
        try:
            ref_chain, old_sha = self.get_refs_container().follow(refname)
        except NotBranchError:
            actual_refname = refname
            old_sha = None
        else:
            if ref_chain:
                actual_refname = ref_chain[-1]
            else:
                actual_refname = refname
        if isinstance(source, GitBranch) and lossy:
            raise errors.LossyPushToSameVCS(source.controldir, self)
        source_store = get_object_store(source.repository)
        fetch_tags = source.get_config_stack().get("branch.fetch_tags")

        def get_changed_refs(remote_refs):
            if self._refs is not None:
                update_refs_container(self._refs, remote_refs)
            ret = {}
            # TODO(jelmer): Unpeel if necessary
            push_result.new_original_revid = revision_id
            if lossy:
                new_sha = source_store._lookup_revision_sha1(revision_id)
            else:
                try:
                    new_sha = repo.lookup_bzr_revision_id(revision_id)[0]
                except errors.NoSuchRevision:
                    raise errors.NoRoundtrippingSupport(
                        source, self.open_branch(name=name, nascent_ok=True)
                    )
            old_sha = remote_refs.get(actual_refname)
            if not overwrite:
                if remote_divergence(old_sha, new_sha, source_store):
                    raise DivergedBranches(
                        source, self.open_branch(name, nascent_ok=True)
                    )
            ret[actual_refname] = new_sha
            if fetch_tags:
                for tagname, revid in source.tags.get_tag_dict().items():
                    if tag_selector and not tag_selector(tagname):
                        continue
                    if lossy:
                        try:
                            new_sha = source_store._lookup_revision_sha1(revid)
                        except KeyError:
                            if source.repository.has_revision(revid):
                                raise
                    else:
                        try:
                            new_sha = repo.lookup_bzr_revision_id(revid)[0]
                        except errors.NoSuchRevision:
                            continue
                        else:
                            if not source.repository.has_revision(revid):
                                continue
                    ret[tag_name_to_ref(tagname)] = new_sha
            return ret

        with source_store.lock_read():

            def generate_pack_data(have, want, progress=None, ofs_delta=True):
                git_repo = getattr(source.repository, "_git", None)
                if git_repo:
                    shallow = git_repo.get_shallow()
                else:
                    shallow = None
                if lossy:
                    return source_store.generate_lossy_pack_data(
                        have,
                        want,
                        shallow=shallow,
                        progress=progress,
                        ofs_delta=ofs_delta,
                    )
                elif shallow:
                    return source_store.generate_pack_data(
                        have,
                        want,
                        shallow=shallow,
                        progress=progress,
                        ofs_delta=ofs_delta,
                    )
                else:
                    return source_store.generate_pack_data(
                        have, want, progress=progress, ofs_delta=ofs_delta
                    )

            dw_result = self.send_pack(get_changed_refs, generate_pack_data)
            new_refs = dw_result.refs
            error = dw_result.ref_status.get(actual_refname)
            if error:
                raise error
            for ref, error in dw_result.ref_status.items():
                if error:
                    trace.warning("unable to open ref %s: %s", ref, error)
        push_result.new_revid = repo.lookup_foreign_revision_id(
            new_refs[actual_refname]
        )
        if old_sha is not None:
            push_result.old_revid = repo.lookup_foreign_revision_id(old_sha)
        else:
            push_result.old_revid = NULL_REVISION
        if self._refs is not None:
            update_refs_container(self._refs, new_refs)
        push_result.target_branch = self.open_branch(name)
        if old_sha is not None:
            push_result.branch_push_result = GitBranchPushResult()
            push_result.branch_push_result.source_branch = source
            push_result.branch_push_result.target_branch = push_result.target_branch
            push_result.branch_push_result.local_branch = None
            push_result.branch_push_result.master_branch = push_result.target_branch
            push_result.branch_push_result.old_revid = push_result.old_revid
            push_result.branch_push_result.new_revid = push_result.new_revid
            push_result.branch_push_result.new_original_revid = (
                push_result.new_original_revid
            )
        if source.get_push_location() is None or remember:
            source.set_push_location(push_result.target_branch.base)
        return push_result

    def _find_commondir(self):
        # There is no way to find the commondir, if there is any.
        return self


class EmptyObjectStoreIterator(dict):
    def iterobjects(self):
        return []


class TemporaryPackIterator(Pack):
    def __init__(self, path, resolve_ext_ref):
        super().__init__(path, resolve_ext_ref=resolve_ext_ref)
        self._idx_load = lambda: self._idx_load_or_generate(self._idx_path)

    def _idx_load_or_generate(self, path):
        if not os.path.exists(path):
            with ui.ui_factory.nested_progress_bar() as pb:

                def report_progress(cur, total):
                    pb.update("generating index", cur, total)

                self.data.create_index(path, progress=report_progress)
        return load_pack_index(path)

    def __del__(self):
        if self._idx is not None:
            self._idx.close()
            os.remove(self._idx_path)
        if self._data is not None:
            self._data.close()
            os.remove(self._data_path)


class BzrGitHttpClient(dulwich.client.HttpGitClient):
    def __init__(self, transport, *args, **kwargs):
        self.transport = transport
        url = urlutils.URL.from_string(transport.external_url())
        url.user = url.quoted_user = None
        url.password = url.quoted_password = None
        url = urlutils.strip_segment_parameters(str(url))
        super().__init__(url, *args, **kwargs)

    def archive(
        self,
        path,
        committish,
        write_data,
        progress=None,
        write_error=None,
        format=None,
        subdirs=None,
        prefix=None,
    ):
        raise GitSmartRemoteNotSupported(self.archive, self)

    def _http_request(self, url, headers=None, data=None, allow_compression=False):
        """Perform HTTP request.

        :param url: Request URL.
        :param headers: Optional custom headers to override defaults.
        :param data: Request data.
        :return: Tuple (`response`, `read`), where response is an `urllib3`
            response object with additional `content_type` and
            `redirect_location` properties, and `read` is a consumable read
            method for the response data.
        """
        if is_github_url(url):
            headers["User-agent"] = user_agent_for_github()
        headers["Pragma"] = "no-cache"

        response = self.transport.request(
            ("GET" if data is None else "POST"),
            url,
            body=data,
            headers=headers,
            retries=8,
        )

        if response.status == 404:
            raise NotGitRepository()
        elif response.status != 200:
            raise GitProtocolError(
                f"unexpected http resp {response.status} for {url}",
            )

        read = response.read

        class WrapResponse:
            def __init__(self, response):
                self._response = response
                self.status = response.status
                self.content_type = response.getheader("Content-Type")
                self.redirect_location = response._actual.geturl()

            def readlines(self):
                return self._response.readlines()

            def close(self):
                pass

        return WrapResponse(response), read


def _git_url_and_path_from_transport(external_url):
    url = urlutils.strip_segment_parameters(external_url)
    return urlparse.urlsplit(url)


class RemoteGitControlDirFormat(GitControlDirFormat):
    """The .git directory control format."""

    supports_workingtrees = False

    @classmethod
    def _known_formats(self):
        return {RemoteGitControlDirFormat()}

    def get_branch_format(self):
        return RemoteGitBranchFormat()

    @property
    def repository_format(self):
        return GitRepositoryFormat()

    def is_initializable(self):
        return False

    def is_supported(self):
        return True

    def open(self, transport, _found=None):
        """Open this directory."""
        split_url = _git_url_and_path_from_transport(transport.external_url())
        if isinstance(transport, GitSmartTransport):
            client = transport._get_client()
        elif split_url.scheme in ("http", "https"):
            client = BzrGitHttpClient(transport)
        elif split_url.scheme in ("file",):
            client = dulwich.client.LocalGitClient()
        else:
            raise NotBranchError(transport.base)
        if not _found:
            pass  # TODO(jelmer): Actually probe for something
        return RemoteGitDir(transport, self, client, split_url.path)

    def get_format_description(self):
        return "Remote Git Repository"

    def initialize_on_transport(self, transport):
        raise UninitializableFormat(self)

    def supports_transport(self, transport):
        try:
            external_url = transport.external_url()
        except InProcessTransport:
            raise NotBranchError(path=transport.base)
        return (
            external_url.startswith("http:")
            or external_url.startswith("https:")
            or external_url.startswith("git+")
            or external_url.startswith("git:")
        )


class GitRemoteRevisionTree(RevisionTree):
    def archive(
        self,
        format,
        name,
        root=None,
        subdir=None,
        force_mtime=None,
        recurse_nested=False,
    ):
        """Create an archive of this tree.

        :param format: Format name (e.g. 'tar')
        :param name: target file name
        :param root: Root directory name (or None)
        :param subdir: Subdirectory to export (or None)
        :return: Iterator over archive chunks
        """
        if recurse_nested:
            # TODO(jelmer): Parse .gitmodules from archive afterwards?
            raise NotImplementedError("recurse_nested is not yet supported")
        commit = self._repository.lookup_bzr_revision_id(self.get_revision_id())[0]
        from tempfile import SpooledTemporaryFile

        f = SpooledTemporaryFile(max_size=PACK_SPOOL_FILE_MAX_SIZE, prefix="incoming-")
        # git-upload-archive(1) generaly only supports refs. So let's see if we
        # can find one.
        reverse_refs = {
            v: k
            for (k, v) in self._repository.controldir.get_refs_container()
            .as_dict()
            .items()
        }
        try:
            committish = reverse_refs[commit]
        except KeyError:
            # No? Maybe the user has uploadArchive.allowUnreachable enabled.
            # Let's hope for the best.
            committish = commit
        self._repository.archive(
            format,
            committish,
            f.write,
            subdirs=([subdir] if subdir else None),
            prefix=(root + "/") if root else "",
        )
        f.seek(0)
        return osutils.file_iterator(f)

    def is_versioned(self, path):
        raise GitSmartRemoteNotSupported(self.is_versioned, self)

    def has_filename(self, path):
        raise GitSmartRemoteNotSupported(self.has_filename, self)

    def get_file_text(self, path):
        raise GitSmartRemoteNotSupported(self.get_file_text, self)

    def list_files(self, include_root=False, from_dir=None, recursive=True):
        raise GitSmartRemoteNotSupported(self.list_files, self)


class RemoteGitRepository(GitRepository):
    supports_random_access = False

    @property
    def user_url(self):
        return self.control_url

    def get_parent_map(self, revids):
        raise GitSmartRemoteNotSupported(self.get_parent_map, self)

    def archive(self, *args, **kwargs):
        return self.controldir.archive(*args, **kwargs)

    def fetch_pack(self, determine_wants, graph_walker, pack_data, progress=None):
        return self.controldir.fetch_pack(
            determine_wants, graph_walker, pack_data, progress
        )

    def send_pack(self, get_changed_refs, generate_pack_data):
        return self.controldir.send_pack(get_changed_refs, generate_pack_data)

    def fetch_objects(
        self, determine_wants, graph_walker, resolve_ext_ref, progress=None
    ):
        import tempfile

        fd, path = tempfile.mkstemp(suffix=".pack")
        try:
            self.fetch_pack(
                determine_wants, graph_walker, lambda x: os.write(fd, x), progress
            )
        finally:
            os.close(fd)
        if os.path.getsize(path) == 0:
            return EmptyObjectStoreIterator()
        return TemporaryPackIterator(path[: -len(".pack")], resolve_ext_ref)

    def lookup_bzr_revision_id(self, bzr_revid, mapping=None):
        # This won't work for any round-tripped bzr revisions, but it's a
        # start..
        try:
            return mapping_registry.revision_id_bzr_to_foreign(bzr_revid)
        except InvalidRevisionId:
            raise NoSuchRevision(self, bzr_revid)

    def lookup_foreign_revision_id(self, foreign_revid, mapping=None):
        """Lookup a revision id."""
        if mapping is None:
            mapping = self.get_mapping()
        # Not really an easy way to parse foreign revids here..
        return mapping.revision_id_foreign_to_bzr(foreign_revid)

    def revision_tree(self, revid):
        return GitRemoteRevisionTree(self, revid)

    def get_revisions(self, revids):
        raise GitSmartRemoteNotSupported(self.get_revisions, self)

    def has_revisions(self, revids):
        raise GitSmartRemoteNotSupported(self.get_revisions, self)


class RemoteGitTagDict(GitTags):
    def set_tag(self, name, revid):
        sha = self.branch.lookup_bzr_revision_id(revid)[0]
        self._set_ref(name, sha)

    def delete_tag(self, name):
        self._set_ref(name, dulwich.client.ZERO_SHA)

    def _set_ref(self, name, sha):
        ref = tag_name_to_ref(name)

        def get_changed_refs(old_refs):
            ret = {}
            if sha == dulwich.client.ZERO_SHA and ref not in old_refs:
                raise NoSuchTag(name)
            ret[ref] = sha
            return ret

        def generate_pack_data(have, want, ofs_delta=False, progress=None):
            return pack_objects_to_data([])

        result = self.repository.send_pack(get_changed_refs, generate_pack_data)
        error = result.ref_status.get(ref)
        if error:
            raise error


class RemoteGitBranch(GitBranch):
    def __init__(self, controldir, repository, ref, sha):
        self._sha = sha
        super().__init__(controldir, repository, ref, RemoteGitBranchFormat())

    def last_revision_info(self):
        raise GitSmartRemoteNotSupported(self.last_revision_info, self)

    @property
    def user_url(self):
        return self.control_url

    @property
    def control_url(self):
        return self.base

    def revision_id_to_revno(self, revision_id):
        raise GitSmartRemoteNotSupported(self.revision_id_to_revno, self)

    def last_revision(self):
        return self.lookup_foreign_revision_id(self.head)

    @property
    def head(self):
        return self._sha

    def _synchronize_history(self, destination, revision_id):
        """See Branch._synchronize_history()."""
        if revision_id is None:
            revision_id = self.last_revision()
        destination.generate_revision_history(revision_id)

    def _get_parent_location(self):
        return None

    def get_push_location(self):
        return None

    def set_push_location(self, url):
        pass

    def _iter_tag_refs(self):
        """Iterate over the tag refs.

        :param refs: Refs dictionary (name -> git sha1)
        :return: iterator over (ref_name, tag_name, peeled_sha1, unpeeled_sha1)
        """
        refs = self.controldir.get_refs_container()
        for ref_name, unpeeled in refs.as_dict().items():
            try:
                tag_name = ref_to_tag_name(ref_name)
            except (ValueError, UnicodeDecodeError):
                continue
            peeled = refs.get_peeled(ref_name)
            if peeled is None:
                # Let's just hope it's a commit
                peeled = unpeeled
            if not isinstance(tag_name, str):
                raise TypeError(tag_name)
            yield (ref_name, tag_name, peeled, unpeeled)

    def set_last_revision_info(self, revno, revid):
        self.generate_revision_history(revid)

    def generate_revision_history(self, revision_id, last_rev=None, other_branch=None):
        sha = self.lookup_bzr_revision_id(revision_id)[0]

        def get_changed_refs(old_refs):
            return {self.ref: sha}

        def generate_pack_data(have, want, ofs_delta=False, progress=None):
            return pack_objects_to_data([])

        result = self.repository.send_pack(get_changed_refs, generate_pack_data)
        error = result.ref_status.get(self.ref)
        if error:
            raise error
        self._sha = sha


def remote_refs_dict_to_container(refs_dict, symrefs_dict=None):
    if symrefs_dict is None:
        symrefs_dict = {}
    base = {}
    peeled = {}
    for k, v in refs_dict.items():
        if is_peeled(k):
            peeled[k[:-3]] = v
        else:
            base[k] = v
    for name, target in symrefs_dict.items():
        base[name] = SYMREF + target
    ret = DictRefsContainer(base)
    ret._peeled = peeled
    return ret


def update_refs_container(container, refs_dict):
    peeled = {}
    base = {}
    for k, v in refs_dict.items():
        if is_peeled(k):
            peeled[k[:-3]] = v
        else:
            base[k] = v
    container._peeled = peeled
    container._refs.update(base)



# Copyright (C) 2008, 2009, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Views stored within a working tree.

The views are actually in the WorkingTree.views namespace, but these are
1:1 with WorkingTree implementations so can be tested from here.
"""

from breezy import views as _mod_views
from breezy.tests import TestNotApplicable, TestSkipped
from breezy.tests.per_workingtree import TestCaseWithWorkingTree
from breezy.workingtree import WorkingTree


class TestTreeViews(TestCaseWithWorkingTree):
    def setUp(self):
        # formats that don't support views can skip the rest of these
        # tests...
        fmt = self.workingtree_format
        f = fmt.supports_views
        if f is None:
            raise TestSkipped(
                "format {} doesn't declare whether it "
                "supports views, assuming not".format(fmt)
            )
        if not f():
            raise TestNotApplicable("format {} doesn't support views".format(fmt))
        super().setUp()

    def test_views_initially_empty(self):
        wt = self.make_branch_and_tree("wt")
        current, views = wt.views.get_view_info()
        self.assertEqual(None, current)
        self.assertEqual({}, views)

    def test_set_and_get_view_info(self):
        wt = self.make_branch_and_tree("wt")
        view_current = "view-name"
        view_dict = {view_current: ["dir-1"], "other-name": ["dir-2"]}
        wt.views.set_view_info(view_current, view_dict)
        current, views = wt.views.get_view_info()
        self.assertEqual(view_current, current)
        self.assertEqual(view_dict, views)
        # then reopen the tree and see they're still there
        wt = WorkingTree.open("wt")
        current, views = wt.views.get_view_info()
        self.assertEqual(view_current, current)
        self.assertEqual(view_dict, views)
        # test setting a current view which does not exist
        self.assertRaises(
            _mod_views.NoSuchView, wt.views.set_view_info, "yet-another", view_dict
        )
        current, views = wt.views.get_view_info()
        self.assertEqual(view_current, current)
        self.assertEqual(view_dict, views)
        # test clearing the current view
        wt.views.set_view_info(None, view_dict)
        current, views = wt.views.get_view_info()
        self.assertEqual(None, current)
        self.assertEqual(view_dict, views)

    def test_lookup_view(self):
        wt = self.make_branch_and_tree("wt")
        view_current = "view-name"
        view_dict = {view_current: ["dir-1"], "other-name": ["dir-2"]}
        wt.views.set_view_info(view_current, view_dict)
        # test lookup of the default view
        result = wt.views.lookup_view()
        self.assertEqual(result, ["dir-1"])
        # test lookup of a named view
        result = wt.views.lookup_view("other-name")
        self.assertEqual(result, ["dir-2"])

    def test_set_view(self):
        wt = self.make_branch_and_tree("wt")
        # test that set_view sets the current view by default
        wt.views.set_view("view-1", ["dir-1"])
        current, views = wt.views.get_view_info()
        self.assertEqual("view-1", current)
        self.assertEqual({"view-1": ["dir-1"]}, views)
        # test adding a view and not making it the current one
        wt.views.set_view("view-2", ["dir-2"], make_current=False)
        current, views = wt.views.get_view_info()
        self.assertEqual("view-1", current)
        self.assertEqual({"view-1": ["dir-1"], "view-2": ["dir-2"]}, views)

    def test_unicode_view(self):
        wt = self.make_branch_and_tree("wt")
        view_name = "\u3070"
        view_files = ["foo", "bar/"]
        view_dict = {view_name: view_files}
        wt.views.set_view_info(view_name, view_dict)
        current, views = wt.views.get_view_info()
        self.assertEqual(view_name, current)
        self.assertEqual(view_dict, views)

    def test_no_such_view(self):
        wt = self.make_branch_and_tree("wt")
        try:
            wt.views.lookup_view("opaque")
        except _mod_views.NoSuchView as e:
            self.assertEqual(e.view_name, "opaque")
            self.assertEqual(str(e), "No such view: opaque.")
        else:
            self.fail("didn't get expected exception")

    def test_delete_view(self):
        wt = self.make_branch_and_tree("wt")
        view_name = "\N{GREEK SMALL LETTER ALPHA}"
        view_files = ["alphas/"]
        wt.views.set_view(view_name, view_files)
        # now try to delete it
        wt.views.delete_view(view_name)
        # now you can't look it up
        self.assertRaises(_mod_views.NoSuchView, wt.views.lookup_view, view_name)
        # and it's not in the dictionary
        self.assertEqual(wt.views.get_view_info()[1], {})
        # and you can't remove it a second time
        self.assertRaises(_mod_views.NoSuchView, wt.views.delete_view, view_name)
        # or remove a view that never existed
        self.assertRaises(_mod_views.NoSuchView, wt.views.delete_view, view_name + "2")

    def test_check_path_in_view(self):
        wt = self.make_branch_and_tree("wt")
        view_current = "view-name"
        view_dict = {view_current: ["dir-1"], "other-name": ["dir-2"]}
        wt.views.set_view_info(view_current, view_dict)
        self.assertEqual(_mod_views.check_path_in_view(wt, "dir-1"), None)
        self.assertEqual(_mod_views.check_path_in_view(wt, "dir-1/sub"), None)
        self.assertRaises(
            _mod_views.FileOutsideView, _mod_views.check_path_in_view, wt, "dir-2"
        )
        self.assertRaises(
            _mod_views.FileOutsideView, _mod_views.check_path_in_view, wt, "dir-2/sub"
        )
        self.assertRaises(
            _mod_views.FileOutsideView, _mod_views.check_path_in_view, wt, "other"
        )


class TestUnsupportedViews(TestCaseWithWorkingTree):
    """Formats that don't support views should give reasonable errors."""

    def setUp(self):
        fmt = self.workingtree_format
        supported = fmt.supports_views
        if supported is None:
            warn(
                "Format {} doesn't declare whether it supports views or not".format(fmt)
            )
            raise TestSkipped("No view support at all")
        if supported():
            raise TestSkipped("Format {} declares that views are supported".format(fmt))
            # it's covered by TestTreeViews
        super().setUp()

    def test_view_methods_raise(self):
        wt = self.make_branch_and_tree("wt")
        self.assertRaises(
            _mod_views.ViewsNotSupported,
            wt.views.set_view_info,
            "bar",
            {"bar": ["bars/"]},
        )
        self.assertRaises(_mod_views.ViewsNotSupported, wt.views.get_view_info)
        self.assertRaises(_mod_views.ViewsNotSupported, wt.views.lookup_view, "foo")
        self.assertRaises(_mod_views.ViewsNotSupported, wt.views.set_view, "foo", "bar")
        self.assertRaises(_mod_views.ViewsNotSupported, wt.views.delete_view, "foo")



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.dirty_tracker."""

import os

import breezy
from breezy.tests import TestCaseWithTransport


class DirtyTrackerTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree("tree")
        try:
            from breezy.dirty_tracker import DirtyTracker
        except ModuleNotFoundError as e:
            if e.name == "pyinotify":
                self.skipTest("pyinotify not available")
            raise
        self.tracker = DirtyTracker(self.tree)

    def test_nothing_changes(self):
        try:
            with self.tracker:
                self.assertFalse(self.tracker.is_dirty())
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_regular_file_added(self):
        try:
            with self.tracker:
                self.build_tree_contents([("tree/foo", "bar")])
                self.assertTrue(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), {"foo"})
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_many_added(self):
        try:
            with self.tracker:
                self.build_tree_contents(
                    [("tree/f%d" % d, "content") for d in range(100)]
                )
                self.assertTrue(self.tracker.is_dirty())
                self.assertEqual(
                    self.tracker.relpaths(), {"f%d" % d for d in range(100)}
                )
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_regular_file_in_subdir_added(self):
        try:
            with self.tracker:
                self.build_tree_contents([("tree/foo/",), ("tree/foo/blah", "bar")])
                self.assertTrue(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), {"foo", "foo/blah"})
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_directory_added(self):
        try:
            with self.tracker:
                self.build_tree_contents([("tree/foo/",)])
                self.assertTrue(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), {"foo"})
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_file_removed(self):
        try:
            with self.tracker:
                self.build_tree_contents([("tree/foo", "foo")])
                self.assertTrue(self.tracker.is_dirty())
                self.tracker.mark_clean()
                self.build_tree_contents([("tree/foo", "bar")])
                self.assertTrue(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), {"foo"})
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_control_file(self):
        try:
            with self.tracker:
                self.tree.commit("Some change")
                self.assertFalse(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), set())
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_renamed(self):
        try:
            with self.tracker:
                self.build_tree_contents([("tree/foo", "bar")])
                self.tracker.mark_clean()
                self.assertFalse(self.tracker.is_dirty())
                os.rename("tree/foo", "tree/bar")
                self.assertTrue(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), {"foo", "bar"})
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_deleted(self):
        try:
            with self.tracker:
                self.build_tree_contents([("tree/foo", "bar")])
                self.tracker.mark_clean()
                self.assertFalse(self.tracker.is_dirty())
                os.unlink("tree/foo")
                self.assertTrue(self.tracker.is_dirty(), self.tracker._paths)
                self.assertEqual(self.tracker.relpaths(), {"foo"})
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_added_then_deleted(self):
        try:
            with self.tracker:
                self.tracker.mark_clean()
                self.assertFalse(self.tracker.is_dirty())
                self.build_tree_contents([("tree/foo", "bar")])
                os.unlink("tree/foo")
                self.assertFalse(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), set())
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")

    def test_file_modified(self):
        self.build_tree_contents([("tree/foo", "bla")])
        try:
            with self.tracker:
                self.assertFalse(self.tracker.is_dirty())
                self.build_tree_contents([("tree/foo", "bar")])
                self.assertTrue(self.tracker.is_dirty())
                self.assertEqual(self.tracker.relpaths(), {"foo"})
        except breezy.dirty_tracker.TooManyOpenFiles:
            self.skipTest("Too many open files")



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for msg editor interactions.."""

from .... import msgeditor
from ....tests import TestCaseWithTransport
from ... import commitfromnews


class TestRegisters(TestCaseWithTransport):
    def test_registered_at_import(self):
        self.assertTrue(commitfromnews._registered)

    def test_register_registers_for_commit_message_template(self):
        commitfromnews._registered = False
        # Registers only within the plugin
        commitfromnews.register()
        self.assertLength(1, msgeditor.hooks["commit_message_template"])
        self.assertTrue(commitfromnews._registered)



# Copyright (C) 2008, 2009, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from io import BytesIO

from .. import filters
from ..filters import (
    ContentFilter,
    ContentFilterContext,
    _get_filter_stack_for,
    _get_registered_names,
    filtered_input_file,
    filtered_output_bytes,
    internal_size_sha_file_byname,
)
from ..osutils import sha_string
from . import TestCase, TestCaseInTempDir


# sample filter stacks
def _swapcase(chunks, context=None):
    return [s.swapcase() for s in chunks]


def _addjunk(chunks):
    return [b"junk\n"] + list(chunks)


def _deljunk(chunks, context):
    return list(chunks[1:])


_stack_1 = [
    ContentFilter(_swapcase, _swapcase),
]
_stack_2 = [
    ContentFilter(_swapcase, _swapcase),
    ContentFilter(_addjunk, _deljunk),
]

# sample data
_sample_external = [b"Hello\n", b"World\n"]
_internal_1 = [b"hELLO\n", b"wORLD\n"]
_internal_2 = [b"junk\n", b"hELLO\n", b"wORLD\n"]


class TestContentFilterContext(TestCase):
    def test_empty_filter_context(self):
        ctx = ContentFilterContext()
        self.assertEqual(None, ctx.relpath())

    def test_filter_context_with_path(self):
        ctx = ContentFilterContext("foo/bar")
        self.assertEqual("foo/bar", ctx.relpath())


class TestFilteredInput(TestCase):
    def test_filtered_input_file(self):
        # test an empty stack returns the same result
        external = b"".join(_sample_external)
        f = BytesIO(external)
        (fileobj, size) = filtered_input_file(f, [])
        self.assertEqual((external, 12), (fileobj.read(), size))
        # test a single item filter stack
        f = BytesIO(external)
        expected = b"".join(_internal_1)
        (fileobj, size) = filtered_input_file(f, _stack_1)
        self.assertEqual((expected, 12), (fileobj.read(), size))
        # test a multi item filter stack
        f = BytesIO(external)
        expected = b"".join(_internal_2)
        (fileobj, size) = filtered_input_file(f, _stack_2)
        self.assertEqual((expected, 17), (fileobj.read(), size))


class TestFilteredOutput(TestCase):
    def test_filtered_output_bytes(self):
        # test an empty stack returns the same result
        self.assertEqual(
            _sample_external, list(filtered_output_bytes(_sample_external, None))
        )
        # test a single item filter stack
        self.assertEqual(
            _sample_external, list(filtered_output_bytes(_internal_1, _stack_1))
        )
        # test a multi item filter stack
        self.assertEqual(
            _sample_external, list(filtered_output_bytes(_internal_2, _stack_2))
        )


class TestFilteredSha(TestCaseInTempDir):
    def test_filtered_size_sha(self):
        # check that the size and sha matches what's expected
        text = b"Foo Bar Baz\n"
        with open("a", "wb") as a:
            a.write(text)
        post_filtered_content = b"".join(_swapcase([text], None))
        expected_len = len(post_filtered_content)
        expected_sha = sha_string(post_filtered_content)
        self.assertEqual(
            (expected_len, expected_sha),
            internal_size_sha_file_byname("a", [ContentFilter(_swapcase, _swapcase)]),
        )


class TestFilterStackMaps(TestCase):
    def _register_map(self, pref, stk1, stk2):
        def stk_lookup(key):
            return {"v1": stk1, "v2": stk2}.get(key)

        filters.filter_stacks_registry.register(pref, stk_lookup)

    def test_filter_stack_maps(self):
        # Save the current registry
        original_registry = filters._reset_registry()
        self.addCleanup(filters._reset_registry, original_registry)
        # Test registration
        a_stack = [ContentFilter("b", "c")]
        z_stack = [ContentFilter("y", "x"), ContentFilter("w", "v")]
        self._register_map("foo", a_stack, z_stack)
        self.assertEqual(["foo"], _get_registered_names())
        self._register_map("bar", z_stack, a_stack)
        self.assertEqual(["bar", "foo"], _get_registered_names())
        # Test re-registration raises an error
        self.assertRaises(KeyError, self._register_map, "foo", [], [])

    def test_get_filter_stack_for(self):
        # Save the current registry
        original_registry = filters._reset_registry()
        self.addCleanup(filters._reset_registry, original_registry)
        # Test filter stack lookup
        a_stack = [ContentFilter("b", "c")]
        d_stack = [ContentFilter("d", "D")]
        z_stack = [ContentFilter("y", "x"), ContentFilter("w", "v")]
        self._register_map("foo", a_stack, z_stack)
        self._register_map("bar", d_stack, z_stack)
        prefs = (("foo", "v1"),)
        self.assertEqual(a_stack, _get_filter_stack_for(prefs))
        prefs = (("foo", "v2"),)
        self.assertEqual(z_stack, _get_filter_stack_for(prefs))
        prefs = (("foo", "v1"), ("bar", "v1"))
        self.assertEqual(a_stack + d_stack, _get_filter_stack_for(prefs))
        # Test an unknown preference
        prefs = (("baz", "v1"),)
        self.assertEqual([], _get_filter_stack_for(prefs))
        # Test an unknown value
        prefs = (("foo", "v3"),)
        self.assertEqual([], _get_filter_stack_for(prefs))
        # Test a value of None is skipped
        prefs = (("foo", None), ("bar", "v1"))
        self.assertEqual(d_stack, _get_filter_stack_for(prefs))



# Copyright (C) 2006, 2007, 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz revert."""

import os

import breezy.osutils
from breezy.tests import TestCaseWithTransport
from breezy.trace import mutter
from breezy.workingtree import WorkingTree


class TestRevert(TestCaseWithTransport):
    def _prepare_tree(self):
        self.run_bzr("init")
        self.run_bzr("mkdir dir")

        with open("dir/file", "wb") as f:
            f.write(b"spam")
        self.run_bzr("add dir/file")

        self.run_bzr("commit -m1")

        # modify file
        with open("dir/file", "wb") as f:
            f.write(b"eggs")

        # check status
        self.assertEqual("modified:\n  dir/file\n", self.run_bzr("status")[0])

    def _prepare_rename_mod_tree(self):
        self.build_tree(
            ["a/", "a/b", "a/c", "a/d/", "a/d/e", "f/", "f/g", "f/h", "f/i"]
        )
        self.run_bzr("init")
        self.run_bzr("add")
        self.run_bzr("commit -m 1")
        wt = WorkingTree.open(".")
        wt.rename_one("a/b", "f/b")
        wt.rename_one("a/d/e", "f/e")
        wt.rename_one("a/d", "f/d")
        wt.rename_one("f/g", "a/g")
        wt.rename_one("f/h", "h")
        wt.rename_one("f", "j")

    def helper(self, param=""):
        self._prepare_tree()
        # change dir
        # revert to default revision for file in subdir does work
        os.chdir("dir")
        mutter("cd dir\n")

        self.assertEqual("1\n", self.run_bzr("revno")[0])
        self.run_bzr("revert {} file".format(param))
        with open("file", "rb") as f:
            self.assertEqual(b"spam", f.read())

    def test_revert_in_subdir(self):
        self.helper()

    def test_revert_to_revision_in_subdir(self):
        # test case for bug #29424:
        # revert to specific revision for file in subdir does not work
        self.helper("-r 1")

    def test_revert_in_checkout(self):
        os.mkdir("brach")
        os.chdir("brach")
        self._prepare_tree()
        self.run_bzr("checkout --lightweight . ../sprach")
        self.run_bzr("commit -m more")
        os.chdir("../sprach")
        self.assertEqual("", self.run_bzr("status")[0])
        self.run_bzr("revert")
        self.assertEqual("", self.run_bzr("status")[0])

    def test_revert_dirname(self):
        """Test that revert DIRECTORY does what's expected."""
        self._prepare_rename_mod_tree()
        self.run_bzr("revert a")
        self.assertPathExists("a/b")
        self.assertPathExists("a/d")
        self.assertPathDoesNotExist("a/g")
        self.expectFailure(
            "j is in the delta revert applies because j was renamed too",
            self.assertPathExists,
            "j",
        )
        self.assertPathExists("h")
        self.run_bzr("revert f")
        self.assertPathDoesNotExist("j")
        self.assertPathDoesNotExist("h")
        self.assertPathExists("a/d/e")

    def test_revert_chatter(self):
        self._prepare_rename_mod_tree()
        chatter = self.run_bzr("revert")[1]
        self.assertEqualDiff(
            "R   a/g => f/g\n"
            "R   h => f/h\n"
            "R   j/ => f/\n"
            "R   j/b => a/b\n"
            "R   j/d/ => a/d/\n"
            "R   j/e => a/d/e\n",
            chatter,
        )

    def test_revert(self):
        self.run_bzr("init")

        with open("hello", "w") as f:
            f.write("foo")
        self.run_bzr("add hello")
        self.run_bzr("commit -m setup hello")

        with open("goodbye", "w") as f:
            f.write("baz")
        self.run_bzr("add goodbye")
        self.run_bzr("commit -m setup goodbye")

        with open("hello", "w") as f:
            f.write("bar")
        with open("goodbye", "w") as f:
            f.write("qux")
        self.run_bzr("revert hello")
        self.check_file_contents("hello", b"foo")
        self.check_file_contents("goodbye", b"qux")
        self.run_bzr("revert")
        self.check_file_contents("goodbye", b"baz")

        os.mkdir("revertdir")
        self.run_bzr("add revertdir")
        self.run_bzr("commit -m f")
        os.rmdir("revertdir")
        self.run_bzr("revert")

        if breezy.osutils.supports_symlinks(self.test_dir):
            os.symlink("/unlikely/to/exist", "symlink")
            self.run_bzr("add symlink")
            self.run_bzr("commit -m f")
            os.unlink("symlink")
            self.run_bzr("revert")
            self.assertPathExists("symlink")
            os.unlink("symlink")
            os.symlink("a-different-path", "symlink")
            self.run_bzr("revert")
            self.assertEqual("/unlikely/to/exist", os.readlink("symlink"))
        else:
            self.log("skipping revert symlink tests")

        with open("hello", "w") as f:
            f.write("xyz")
        self.run_bzr("commit -m xyz hello")
        self.run_bzr("revert -r 1 hello")
        self.check_file_contents("hello", b"foo")
        self.run_bzr("revert hello")
        self.check_file_contents("hello", b"xyz")
        os.chdir("revertdir")
        self.run_bzr("revert")
        os.chdir("..")

    def test_revert_newly_added(self):
        # this tests the UI reports reverting a newly added file
        # correct (such files are not deleted)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["file"])
        tree.add(["file"])
        out, err = self.run_bzr("revert")
        self.assertEqual("", out)
        self.assertEqual("-   file\n", err)

    def test_revert_removing_file(self):
        # this tests the UI reports reverting a file which has been committed
        # to a revision that did not have it, reports it as being deleted.
        tree = self.make_branch_and_tree(".")
        tree.commit("empty commit")
        self.build_tree(["file"])
        tree.add(["file"])
        tree.commit("add file")
        out, err = self.run_bzr("revert -r -2")
        self.assertEqual("", out)
        self.assertEqual("-D  file\n", err)

    def test_revert_forget_merges(self):
        # revert --forget-merges removes any pending merges into the tree, but
        # leaves the files unchanged
        tree = self.make_branch_and_tree(".")
        # forget-merges before first commit, though pointless, does not fail
        self.run_bzr(["revert", "--forget-merges"])
        self.build_tree(["file"])
        first_rev_id = tree.commit("initial commit")
        self.build_tree_contents([("file", b"new content")])
        existing_parents = tree.get_parent_ids()
        self.assertEqual([first_rev_id], existing_parents)
        merged_parents = existing_parents + [b"merged-in-rev"]
        tree.set_parent_ids(merged_parents)
        self.assertEqual(merged_parents, tree.get_parent_ids())
        self.run_bzr(["revert", "--forget-merges"])
        self.assertEqual([first_rev_id], tree.get_parent_ids())
        # changed files are not reverted
        self.assertFileEqual(b"new content", "file")
        # you can give it the path of a tree
        self.run_bzr(["revert", "--forget-merges", tree.abspath(".")])



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Author: Aaron Bentley <aaron.bentley@utoronto.ca>

"""Black-box tests for brz merge."""

import doctest
import os

from testtools import matchers

from breezy import (
    branch,
    controldir,
    merge_directive,
    osutils,
    tests,
    urlutils,
    workingtree,
)
from breezy.bzr import conflicts
from breezy.tests import scenarios, script

load_tests = scenarios.load_tests_apply_scenarios


class TestMerge(tests.TestCaseWithTransport):
    def example_branch(self, path="."):
        tree = self.make_branch_and_tree(path)
        self.build_tree_contents(
            [
                (osutils.pathjoin(path, "hello"), b"foo"),
                (osutils.pathjoin(path, "goodbye"), b"baz"),
            ]
        )
        tree.add("hello")
        tree.commit(message="setup")
        tree.add("goodbye")
        tree.commit(message="setup")
        return tree

    def create_conflicting_branches(self):
        """Create two branches which have overlapping modifications.

        :return: (tree, other_branch) Where merging other_branch causes a file
            conflict.
        """
        builder = self.make_branch_builder("branch")
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("fname", b"f-id", "file", b"a\nb\nc\n")),
            ],
            revision_id=b"rev1",
        )
        builder.build_snapshot(
            [b"rev1"], [("modify", ("fname", b"a\nB\nD\n"))], revision_id=b"rev2other"
        )
        other = builder.get_branch().controldir.sprout("other").open_branch()
        builder.build_snapshot(
            [b"rev1"], [("modify", ("fname", b"a\nB\nC\n"))], revision_id=b"rev2this"
        )
        tree = builder.get_branch().create_checkout("tree", lightweight=True)
        return tree, other

    def test_merge_reprocess(self):
        d = controldir.ControlDir.create_standalone_workingtree(".")
        d.commit("h")
        self.run_bzr("merge . --reprocess --merge-type weave")

    def test_merge(self):
        a_tree = self.example_branch("a")
        ancestor = a_tree.branch.revno()
        b_tree = a_tree.controldir.sprout("b").open_workingtree()
        self.build_tree_contents([("b/goodbye", b"quux")])
        b_tree.commit(message="more u's are always good")

        self.build_tree_contents([("a/hello", b"quuux")])
        # We can't merge when there are in-tree changes
        self.run_bzr("merge ../b", retcode=3, working_dir="a")
        a = workingtree.WorkingTree.open("a")
        a_tip = a.commit("Like an epidemic of u's")

        def run_merge_then_revert(args, retcode=None, working_dir="a"):
            self.run_bzr(
                ["merge", "../b", "-r", "last:1..last:1"] + args,
                retcode=retcode,
                working_dir=working_dir,
            )
            if retcode != 3:
                a_tree.revert(backups=False)

        run_merge_then_revert(["--merge-type", "bloof"], retcode=3)
        run_merge_then_revert(["--merge-type", "merge3"])
        run_merge_then_revert(["--merge-type", "weave"])
        run_merge_then_revert(["--merge-type", "lca"])
        self.run_bzr_error(
            ["Show-base is not supported for this merge type"],
            "merge ../b -r last:1..last:1 --merge-type weave --show-base",
            working_dir="a",
        )
        a_tree.revert(backups=False)
        self.run_bzr("merge ../b -r last:1..last:1 --reprocess", working_dir="a")
        a_tree.revert(backups=False)
        self.run_bzr("merge ../b -r last:1", working_dir="a")
        self.check_file_contents("a/goodbye", b"quux")
        # Merging a branch pulls its revision into the tree
        b = branch.Branch.open("b")
        b_tip = b.last_revision()
        self.assertTrue(a.branch.repository.has_revision(b_tip))
        self.assertEqual([a_tip, b_tip], a.get_parent_ids())
        a_tree.revert(backups=False)
        out, err = self.run_bzr("merge -r revno:1:./hello", retcode=3, working_dir="a")
        self.assertTrue("Not a branch" in err)
        self.run_bzr(
            "merge -r revno:%d:./..revno:%d:../b" % (ancestor, b.revno()),
            working_dir="a",
        )
        self.assertEqual(
            a.get_parent_ids(), [a.branch.last_revision(), b.last_revision()]
        )
        self.check_file_contents("a/goodbye", b"quux")
        a_tree.revert(backups=False)
        self.run_bzr("merge -r revno:%d:../b" % b.revno(), working_dir="a")
        self.assertEqual(
            a.get_parent_ids(), [a.branch.last_revision(), b.last_revision()]
        )
        a_tip = a.commit("merged")
        self.run_bzr("merge ../b -r last:1", working_dir="a")
        self.assertEqual([a_tip], a.get_parent_ids())

    def test_merge_defaults_to_reprocess(self):
        tree, other = self.create_conflicting_branches()
        # The default merge algorithm should enable 'reprocess' because
        # 'show-base' is not set
        self.run_bzr("merge ../other", working_dir="tree", retcode=1)
        self.assertEqualDiff(
            b"a\nB\n<<<<<<< TREE\nC\n=======\nD\n>>>>>>> MERGE-SOURCE\n",
            tree.get_file_text("fname"),
        )

    def test_merge_explicit_reprocess_show_base(self):
        tree, other = self.create_conflicting_branches()
        # Explicitly setting --reprocess, and --show-base is an error
        self.run_bzr_error(
            ["Cannot do conflict reduction and show base"],
            "merge ../other --reprocess --show-base",
            working_dir="tree",
        )

    def test_merge_override_reprocess(self):
        tree, other = self.create_conflicting_branches()
        # Explicitly disable reprocess
        self.run_bzr("merge ../other --no-reprocess", working_dir="tree", retcode=1)
        self.assertEqualDiff(
            b"a\n<<<<<<< TREE\nB\nC\n=======\nB\nD\n>>>>>>> MERGE-SOURCE\n",
            tree.get_file_text("fname"),
        )

    def test_merge_override_show_base(self):
        tree, other = self.create_conflicting_branches()
        # Setting '--show-base' will auto-disable '--reprocess'
        self.run_bzr("merge ../other --show-base", working_dir="tree", retcode=1)
        self.assertEqualDiff(
            b"a\n"
            b"<<<<<<< TREE\n"
            b"B\n"
            b"C\n"
            b"||||||| BASE-REVISION\n"
            b"b\n"
            b"c\n"
            b"=======\n"
            b"B\n"
            b"D\n"
            b">>>>>>> MERGE-SOURCE\n",
            tree.get_file_text("fname"),
        )

    def test_merge_with_missing_file(self):
        """Merge handles missing file conflicts."""
        self.build_tree_contents(
            [
                ("a/",),
                ("a/sub/",),
                ("a/sub/a.txt", b"hello\n"),
                ("a/b.txt", b"hello\n"),
                ("a/sub/c.txt", b"hello\n"),
            ]
        )
        a_tree = self.make_branch_and_tree("a")
        a_tree.add(["sub", "b.txt", "sub/c.txt", "sub/a.txt"])
        a_tree.commit(message="added a")
        b_tree = a_tree.controldir.sprout("b").open_workingtree()
        self.build_tree_contents(
            [
                ("a/sub/a.txt", b"hello\nthere\n"),
                ("a/b.txt", b"hello\nthere\n"),
                ("a/sub/c.txt", b"hello\nthere\n"),
            ]
        )
        a_tree.commit(message="Added there")
        os.remove("a/sub/a.txt")
        os.remove("a/sub/c.txt")
        os.rmdir("a/sub")
        os.remove("a/b.txt")
        a_tree.commit(message="Removed a.txt")
        self.build_tree_contents(
            [
                ("b/sub/a.txt", b"hello\nsomething\n"),
                ("b/b.txt", b"hello\nsomething\n"),
                ("b/sub/c.txt", b"hello\nsomething\n"),
            ]
        )
        b_tree.commit(message="Modified a.txt")

        self.run_bzr("merge ../a/", retcode=1, working_dir="b")
        self.assertPathExists("b/sub/a.txt.THIS")
        self.assertPathExists("b/sub/a.txt.BASE")

        self.run_bzr("merge ../b/", retcode=1, working_dir="a")
        self.assertPathExists("a/sub/a.txt.OTHER")
        self.assertPathExists("a/sub/a.txt.BASE")

    def test_conflict_leaves_base_this_other_files(self):
        tree, other = self.create_conflicting_branches()
        self.run_bzr("merge ../other", working_dir="tree", retcode=1)
        self.assertFileEqual(b"a\nb\nc\n", "tree/fname.BASE")
        self.assertFileEqual(b"a\nB\nD\n", "tree/fname.OTHER")
        self.assertFileEqual(b"a\nB\nC\n", "tree/fname.THIS")

    def test_weave_conflict_leaves_base_this_other_files(self):
        tree, other = self.create_conflicting_branches()
        self.run_bzr("merge ../other --weave", working_dir="tree", retcode=1)
        self.assertFileEqual(b"a\nb\nc\n", "tree/fname.BASE")
        self.assertFileEqual(b"a\nB\nD\n", "tree/fname.OTHER")
        self.assertFileEqual(b"a\nB\nC\n", "tree/fname.THIS")

    def test_merge_remember(self):
        """Merge changes from one branch to another, test submit location."""
        tree_a = self.make_branch_and_tree("branch_a")
        branch_a = tree_a.branch
        self.build_tree(["branch_a/a"])
        tree_a.add("a")
        tree_a.commit("commit a")
        branch_b = branch_a.controldir.sprout("branch_b").open_branch()
        tree_b = branch_b.controldir.open_workingtree()
        branch_c = branch_a.controldir.sprout("branch_c").open_branch()
        tree_c = branch_c.controldir.open_workingtree()
        self.build_tree(["branch_a/b"])
        tree_a.add("b")
        tree_a.commit("commit b")
        self.build_tree(["branch_c/c"])
        tree_c.add("c")
        tree_c.commit("commit c")
        # reset parent
        parent = branch_b.get_parent()
        branch_b.set_parent(None)
        self.assertEqual(None, branch_b.get_parent())
        # test merge for failure without parent set
        out = self.run_bzr("merge", retcode=3, working_dir="branch_b")
        self.assertEqual(out, ("", "brz: ERROR: No location specified or remembered\n"))

        # test uncommitted changes
        self.build_tree(["branch_b/d"])
        tree_b.add("d")
        self.run_bzr_error(
            ['Working tree ".*" has uncommitted changes'],
            "merge",
            working_dir="branch_b",
        )

        # merge should now pass and implicitly remember merge location
        tree_b.commit("commit d")
        out, err = self.run_bzr("merge ../branch_a", working_dir="branch_b")

        base = urlutils.local_path_from_url(branch_a.base)
        self.assertEndsWith(err, "+N  b\nAll changes applied successfully.\n")
        # re-open branch as external run_brz modified it
        branch_b = branch_b.controldir.open_branch()
        self.assertEqual(
            osutils.abspath(branch_b.get_submit_branch()), osutils.abspath(parent)
        )
        # test implicit --remember when committing new file
        self.build_tree(["branch_b/e"])
        tree_b.add("e")
        tree_b.commit("commit e")
        out, err = self.run_bzr("merge", working_dir="branch_b")
        self.assertStartsWith(
            err, "Merging from remembered submit location {}\n".format(base)
        )
        # re-open tree as external run_brz modified it
        tree_b = branch_b.controldir.open_workingtree()
        tree_b.commit("merge branch_a")
        # test explicit --remember
        out, err = self.run_bzr("merge ../branch_c --remember", working_dir="branch_b")
        self.assertEqual(out, "")
        self.assertEqual(err, "+N  c\nAll changes applied successfully.\n")
        # re-open branch as external run_brz modified it
        branch_b = branch_b.controldir.open_branch()
        self.assertEqual(
            osutils.abspath(branch_b.get_submit_branch()),
            osutils.abspath(branch_c.controldir.root_transport.base),
        )
        # re-open tree as external run_brz modified it
        tree_b = branch_b.controldir.open_workingtree()
        tree_b.commit("merge branch_c")

    def test_merge_bundle(self):
        from breezy.bzr.testament import Testament

        tree_a = self.make_branch_and_tree("branch_a")
        self.build_tree_contents([("branch_a/a", b"hello")])
        tree_a.add("a")
        tree_a.commit("message")

        tree_b = tree_a.controldir.sprout("branch_b").open_workingtree()
        self.build_tree_contents([("branch_a/a", b"hey there")])
        tree_a.commit("message")

        self.build_tree_contents([("branch_b/a", b"goodbye")])
        tree_b.commit("message")
        self.run_bzr("bundle ../branch_a -o ../bundle", working_dir="branch_b")
        self.run_bzr("merge ../bundle", retcode=1, working_dir="branch_a")
        testament_a = Testament.from_revision(
            tree_a.branch.repository, tree_b.get_parent_ids()[0]
        )
        testament_b = Testament.from_revision(
            tree_b.branch.repository, tree_b.get_parent_ids()[0]
        )
        self.assertEqualDiff(testament_a.as_text(), testament_b.as_text())
        tree_a.set_conflicts([])
        tree_a.commit("message")
        # it is legal to attempt to merge an already-merged bundle
        err = self.run_bzr("merge ../bundle", working_dir="branch_a")[1]
        # but it does nothing
        self.assertFalse(tree_a.changes_from(tree_a.basis_tree()).has_changed())
        self.assertEqual("Nothing to do.\n", err)

    def test_merge_uncommitted(self):
        """Check that merge --uncommitted behaves properly."""
        tree_a = self.make_branch_and_tree("a")
        self.build_tree(["a/file_1", "a/file_2"])
        tree_a.add(["file_1", "file_2"])
        tree_a.commit("commit 1")
        tree_b = tree_a.controldir.sprout("b").open_workingtree()
        self.assertPathExists("b/file_1")
        tree_a.rename_one("file_1", "file_i")
        tree_a.commit("commit 2")
        tree_a.rename_one("file_2", "file_ii")
        self.run_bzr("merge a --uncommitted -d b")
        self.assertPathExists("b/file_1")
        self.assertPathExists("b/file_ii")
        tree_b.revert()
        self.run_bzr_error(
            ("Cannot use --uncommitted and --revision",),
            "merge /a --uncommitted -r1 -d b",
        )

    def test_merge_uncommitted_file(self):
        """It should be possible to merge changes from a single file."""
        tree_a = self.make_branch_and_tree("tree_a")
        tree_a.commit("initial commit")
        tree_a.controldir.sprout("tree_b")
        self.build_tree(["tree_a/file1", "tree_a/file2"])
        tree_a.add(["file1", "file2"])
        self.run_bzr(
            ["merge", "--uncommitted", "../tree_a/file1"], working_dir="tree_b"
        )
        self.assertPathExists("tree_b/file1")
        self.assertPathDoesNotExist("tree_b/file2")

    def test_merge_nonexistent_file(self):
        """It should not be possible to merge changes from a file which
        does not exist.
        """
        tree_a = self.make_branch_and_tree("tree_a")
        self.build_tree_contents([("tree_a/file", b"bar\n")])
        tree_a.add(["file"])
        tree_a.commit("commit 1")
        self.run_bzr_error(
            ("Path\\(s\\) do not exist: non/existing",),
            ["merge", "non/existing"],
            working_dir="tree_a",
        )

    def pullable_branch(self):
        tree_a = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/file", b"bar\n")])
        tree_a.add(["file"])
        self.id1 = tree_a.commit("commit 1")

        tree_b = self.make_branch_and_tree("b")
        tree_b.pull(tree_a.branch)
        self.build_tree_contents([("b/file", b"foo\n")])
        self.id2 = tree_b.commit("commit 2")

    def test_merge_pull(self):
        self.pullable_branch()
        (out, err) = self.run_bzr("merge --pull ../b", working_dir="a")
        self.assertContainsRe(out, "Now on revision 2\\.")
        tree_a = workingtree.WorkingTree.open("a")
        self.assertEqual([self.id2], tree_a.get_parent_ids())

    def test_merge_pull_preview(self):
        self.pullable_branch()
        (out, err) = self.run_bzr("merge --pull --preview -d a b")
        self.assertThat(
            out,
            matchers.DocTestMatches(
                """=== modified file 'file'
--- file\t...
+++ file\t...
@@ -1,1 +1,1 @@
-bar
+foo

""",
                doctest.ELLIPSIS | doctest.REPORT_UDIFF,
            ),
        )
        tree_a = workingtree.WorkingTree.open("a")
        self.assertEqual([self.id1], tree_a.get_parent_ids())

    def test_merge_kind_change(self):
        tree_a = self.make_branch_and_tree("tree_a")
        self.build_tree_contents([("tree_a/file", b"content_1")])
        tree_a.add("file", ids=b"file-id")
        tree_a.commit("added file")
        tree_b = tree_a.controldir.sprout("tree_b").open_workingtree()
        os.unlink("tree_a/file")
        self.build_tree(["tree_a/file/"])
        tree_a.commit("changed file to directory")
        self.run_bzr("merge ../tree_a", working_dir="tree_b")
        self.assertEqual("directory", osutils.file_kind("tree_b/file"))
        tree_b.revert()
        self.assertEqual("file", osutils.file_kind("tree_b/file"))
        self.build_tree_contents([("tree_b/file", b"content_2")])
        tree_b.commit("content change")
        self.run_bzr("merge ../tree_a", retcode=1, working_dir="tree_b")
        self.assertEqual(
            tree_b.conflicts(), [conflicts.ContentsConflict("file", file_id="file-id")]
        )

    def test_directive_cherrypick(self):
        source = self.make_branch_and_tree("source")
        source.commit("nothing")
        # see https://bugs.launchpad.net/bzr/+bug/409688 - trying to
        # cherrypick from one branch into another unrelated branch with a
        # different root id will give shape conflicts.  as a workaround we
        # make sure they share the same root id.
        target = source.controldir.sprout("target").open_workingtree()
        self.build_tree(["source/a"])
        source.add("a")
        source.commit("Added a", rev_id=b"rev1")
        self.build_tree(["source/b"])
        source.add("b")
        source.commit("Added b", rev_id=b"rev2")
        target.commit("empty commit")
        self.write_directive("directive", source.branch, "target", b"rev2", b"rev1")
        out, err = self.run_bzr("merge -d target directive")
        self.assertPathDoesNotExist("target/a")
        self.assertPathExists("target/b")
        self.assertContainsRe(err, "Performing cherrypick")

    def write_directive(
        self,
        filename,
        source,
        target,
        revision_id,
        base_revision_id=None,
        mangle_patch=False,
    ):
        md = merge_directive.MergeDirective2.from_objects(
            source.repository,
            revision_id,
            0,
            0,
            target,
            base_revision_id=base_revision_id,
        )
        if mangle_patch:
            md.patch = b"asdf\n"
        self.build_tree_contents([(filename, b"".join(md.to_lines()))])

    def test_directive_verify_warning(self):
        source = self.make_branch_and_tree("source")
        self.build_tree(["source/a"])
        source.add("a")
        source.commit("Added a", rev_id=b"rev1")
        target = self.make_branch_and_tree("target")
        target.commit("empty commit")
        self.write_directive("directive", source.branch, "target", b"rev1")
        err = self.run_bzr("merge -d target directive")[1]
        self.assertNotContainsRe(err, "Preview patch does not match changes")
        target.revert()
        self.write_directive(
            "directive", source.branch, "target", b"rev1", mangle_patch=True
        )
        err = self.run_bzr("merge -d target directive")[1]
        self.assertContainsRe(err, "Preview patch does not match changes")

    def test_merge_arbitrary(self):
        target = self.make_branch_and_tree("target")
        target.commit("empty")
        # We need a revision that has no integer revno
        branch_a = target.controldir.sprout("branch_a").open_workingtree()
        self.build_tree(["branch_a/file1"])
        branch_a.add("file1")
        branch_a.commit("added file1", rev_id=b"rev2a")
        branch_b = target.controldir.sprout("branch_b").open_workingtree()
        self.build_tree(["branch_b/file2"])
        branch_b.add("file2")
        branch_b.commit("added file2", rev_id=b"rev2b")
        branch_b.merge_from_branch(branch_a.branch)
        self.assertPathExists("branch_b/file1")
        branch_b.commit("merged branch_a", rev_id=b"rev3b")

        # It works if the revid has an interger revno
        self.run_bzr("merge -d target -r revid:rev2a branch_a")
        self.assertPathExists("target/file1")
        self.assertPathDoesNotExist("target/file2")
        target.revert()

        # It should work if the revid has no integer revno
        self.run_bzr("merge -d target -r revid:rev2a branch_b")
        self.assertPathExists("target/file1")
        self.assertPathDoesNotExist("target/file2")

    def assertDirectoryContent(self, directory, entries, message=""):
        """Assert whether entries (file or directories) exist in a directory.

        It also checks that there are no extra entries.
        """
        ondisk = os.listdir(directory)
        if set(ondisk) == set(entries):
            return
        if message:
            message += "\n"
        raise AssertionError(
            '{}"{}" directory content is different:\na = {}\nb = {}\n'.format(
                message, directory, sorted(entries), sorted(ondisk)
            )
        )

    def test_cherrypicking_merge(self):
        # make source branch
        source = self.make_branch_and_tree("source")
        for f in ("a", "b", "c", "d"):
            self.build_tree(["source/" + f])
            source.add(f)
            source.commit("added " + f, rev_id=b"rev_" + f.encode("ascii"))
        # target branch
        target = source.controldir.sprout("target", b"rev_a").open_workingtree()
        self.assertDirectoryContent("target", [".bzr", "a"])
        # pick 1 revision
        self.run_bzr("merge -d target -r revid:rev_b..revid:rev_c source")
        self.assertDirectoryContent("target", [".bzr", "a", "c"])
        target.revert()
        # pick 2 revisions
        self.run_bzr("merge -d target -r revid:rev_b..revid:rev_d source")
        self.assertDirectoryContent("target", [".bzr", "a", "c", "d"])
        target.revert()
        # pick 1 revision with option --changes
        self.run_bzr("merge -d target -c revid:rev_d source")
        self.assertDirectoryContent("target", [".bzr", "a", "d"])

    def test_merge_criss_cross(self):
        tree_a = self.make_branch_and_tree("a")
        tree_a.commit("", rev_id=b"rev1")
        tree_b = tree_a.controldir.sprout("b").open_workingtree()
        tree_a.commit("", rev_id=b"rev2a")
        tree_b.commit("", rev_id=b"rev2b")
        tree_a.merge_from_branch(tree_b.branch)
        tree_b.merge_from_branch(tree_a.branch)
        tree_a.commit("", rev_id=b"rev3a")
        tree_b.commit("", rev_id=b"rev3b")
        tree_a.branch.repository.get_graph(tree_b.branch.repository)
        out, err = self.run_bzr(["merge", "-d", "a", "b"])
        self.assertContainsRe(err, "Warning: criss-cross merge encountered.")

    def test_merge_from_submit(self):
        tree_a = self.make_branch_and_tree("a")
        tree_a.commit("test")
        tree_b = tree_a.controldir.sprout("b").open_workingtree()
        tree_c = tree_a.controldir.sprout("c").open_workingtree()
        out, err = self.run_bzr(["merge", "-d", "c"])
        self.assertContainsRe(err, "Merging from remembered parent location .*a\\/")
        with tree_c.branch.lock_write():
            tree_c.branch.set_submit_branch(tree_b.controldir.root_transport.base)
        out, err = self.run_bzr(["merge", "-d", "c"])
        self.assertContainsRe(err, "Merging from remembered submit location .*b\\/")

    def test_remember_sets_submit(self):
        tree_a = self.make_branch_and_tree("a")
        tree_a.commit("rev1")
        tree_b = tree_a.controldir.sprout("b").open_workingtree()
        self.assertIs(tree_b.branch.get_submit_branch(), None)

        # Remember should not happen if using default from parent
        out, err = self.run_bzr(["merge", "-d", "b"])
        refreshed = workingtree.WorkingTree.open("b")
        self.assertIs(refreshed.branch.get_submit_branch(), None)

        # Remember should happen if user supplies location
        out, err = self.run_bzr(["merge", "-d", "b", "a"])
        refreshed = workingtree.WorkingTree.open("b")
        self.assertEqual(
            refreshed.branch.get_submit_branch(), tree_a.controldir.root_transport.base
        )

    def test_no_remember_dont_set_submit(self):
        tree_a = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/file", b"a\n")])
        tree_a.add("file")
        tree_a.commit("rev1")
        tree_b = tree_a.controldir.sprout("b").open_workingtree()
        self.assertIs(tree_b.branch.get_submit_branch(), None)

        # Remember should not happen if using default from parent
        out, err = self.run_bzr(["merge", "-d", "b", "--no-remember"])
        self.assertEqual(None, tree_b.branch.get_submit_branch())

        # Remember should not happen if user supplies location but ask for not
        # remembering it
        out, err = self.run_bzr(["merge", "-d", "b", "--no-remember", "a"])
        self.assertEqual(None, tree_b.branch.get_submit_branch())

    def test_weave_cherrypick(self):
        this_tree = self.make_branch_and_tree("this")
        self.build_tree_contents([("this/file", b"a\n")])
        this_tree.add("file")
        this_tree.commit("rev1")
        other_tree = this_tree.controldir.sprout("other").open_workingtree()
        self.build_tree_contents([("other/file", b"a\nb\n")])
        other_tree.commit("rev2b")
        self.build_tree_contents([("other/file", b"c\na\nb\n")])
        other_tree.commit("rev3b")
        self.run_bzr("merge --weave -d this other -r -2..-1")
        self.assertFileEqual(b"c\na\n", "this/file")

    def test_lca_merge_criss_cross(self):
        tree_a = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/file", b"base-contents\n")])
        tree_a.add("file")
        tree_a.commit("", rev_id=b"rev1")
        tree_b = tree_a.controldir.sprout("b").open_workingtree()
        self.build_tree_contents([("a/file", b"base-contents\nthis-contents\n")])
        tree_a.commit("", rev_id=b"rev2a")
        self.build_tree_contents([("b/file", b"base-contents\nother-contents\n")])
        tree_b.commit("", rev_id=b"rev2b")
        tree_a.merge_from_branch(tree_b.branch)
        self.build_tree_contents([("a/file", b"base-contents\nthis-contents\n")])
        tree_a.set_conflicts([])
        tree_b.merge_from_branch(tree_a.branch)
        self.build_tree_contents([("b/file", b"base-contents\nother-contents\n")])
        tree_b.set_conflicts([])
        tree_a.commit("", rev_id=b"rev3a")
        tree_b.commit("", rev_id=b"rev3b")
        out, err = self.run_bzr(["merge", "-d", "a", "b", "--lca"], retcode=1)
        self.assertFileEqual(
            b"base-contents\n<<<<<<< TREE\nthis-contents\n"
            b"=======\nother-contents\n>>>>>>> MERGE-SOURCE\n",
            "a/file",
        )

    def test_merge_preview(self):
        this_tree = self.make_branch_and_tree("this")
        this_tree.commit("rev1")
        other_tree = this_tree.controldir.sprout("other").open_workingtree()
        self.build_tree_contents([("other/file", b"new line")])
        other_tree.add("file")
        other_tree.commit("rev2a")
        this_tree.commit("rev2b")
        out, err = self.run_bzr(["merge", "-d", "this", "other", "--preview"])
        self.assertContainsRe(out, "\\+new line")
        self.assertNotContainsRe(err, "\\+N  file\n")
        this_tree.lock_read()
        self.addCleanup(this_tree.unlock)
        self.assertEqual([], list(this_tree.iter_changes(this_tree.basis_tree())))

    def test_merge_missing_second_revision_spec(self):
        """Merge uses branch basis when the second revision is unspecified."""
        this = self.make_branch_and_tree("this")
        this.commit("rev1")
        other = self.make_branch_and_tree("other")
        self.build_tree(["other/other_file"])
        other.add("other_file")
        other.commit("rev1b")
        self.run_bzr("merge -d this other -r0..")
        self.assertPathExists("this/other_file")

    def test_merge_interactive_unlocks_branch(self):
        this = self.make_branch_and_tree("this")
        this.commit("empty commit")
        other = this.controldir.sprout("other").open_workingtree()
        other.commit("empty commit 2")
        self.run_bzr("merge -i -d this other")
        this.lock_write()
        this.unlock()

    def test_merge_fetches_tags(self):
        """Tags are updated by merge, and revisions named in those tags are
        fetched.
        """
        # Make a source, sprout a target off it
        builder = self.make_branch_builder("source")
        builder.build_commit(message="Rev 1", rev_id=b"rev-1")
        source = builder.get_branch()
        target_bzrdir = source.controldir.sprout("target")
        # Add a non-ancestry tag to source
        builder.build_commit(message="Rev 2a", rev_id=b"rev-2a")
        source.tags.set_tag("tag-a", b"rev-2a")
        source.set_last_revision_info(1, b"rev-1")
        source.get_config_stack().set("branch.fetch_tags", True)
        builder.build_commit(message="Rev 2b", rev_id=b"rev-2b")
        # Merge from source
        self.run_bzr("merge -d target source")
        target = target_bzrdir.open_branch()
        # The tag is present, and so is its revision.
        self.assertEqual(b"rev-2a", target.tags.lookup_tag("tag-a"))
        target.repository.get_revision(b"rev-2a")


class TestMergeRevisionRange(tests.TestCaseWithTransport):
    scenarios = (("whole-tree", {"context": "."}), ("file-only", {"context": "a"}))

    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree(".")
        self.tree.commit("initial commit")
        for f in ("a", "b"):
            self.build_tree([f])
            self.tree.add(f)
            self.tree.commit("added " + f)

    def test_merge_reversed_revision_range(self):
        self.run_bzr("merge -r 2..1 " + self.context)
        self.assertPathDoesNotExist("a")
        self.assertPathExists("b")


class TestMergeScript(script.TestCaseWithTransportAndScript):
    def test_merge_empty_branch(self):
        source = self.make_branch_and_tree("source")
        self.build_tree(["source/a"])
        source.add("a")
        source.commit("Added a", rev_id=b"rev1")
        self.make_branch_and_tree("target")
        self.run_script("""\
$ brz merge -d target source
2>brz: ERROR: Merging into empty branches not currently supported, https://bugs.launchpad.net/bzr/+bug/308562
""")


class TestMergeForce(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree_a = self.make_branch_and_tree("a")
        self.build_tree(["a/foo"])
        self.tree_a.add(["foo"])
        self.tree_a.commit("add file")
        self.tree_b = self.tree_a.controldir.sprout("b").open_workingtree()
        self.build_tree_contents([("a/foo", b"change 1")])
        self.tree_a.commit("change file")
        self.tree_b.merge_from_branch(self.tree_a.branch)

    def test_merge_force(self):
        self.tree_a.commit("empty change to allow merge to run")
        # Second merge on top of the uncommitted one
        self.run_bzr(["merge", "../a", "--force"], working_dir="b")

    def test_merge_with_uncommitted_changes(self):
        self.run_bzr_error(
            ["Working tree .* has uncommitted changes"],
            ["merge", "../a"],
            working_dir="b",
        )

    def test_merge_with_pending_merges(self):
        # Revert the changes keeping the pending merge
        self.run_bzr(["revert", "b"])
        self.run_bzr_error(
            ["Working tree .* has uncommitted changes"],
            ["merge", "../a"],
            working_dir="b",
        )



# Copyright (C) 2005-2009, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for topological sort."""

import pprint

from breezy.errors import GraphCycleError
from breezy.revision import NULL_REVISION
from breezy.tests import TestCase
from breezy.tsort import MergeSorter, TopoSorter, merge_sort, topo_sort


class TopoSortTests(TestCase):
    def assertSortAndIterate(self, graph, result_list):
        """Check that sorting and iter_topo_order on graph works."""
        self.assertEqual(result_list, topo_sort(graph))
        self.assertEqual(result_list, list(TopoSorter(graph).iter_topo_order()))

    def assertSortAndIterateRaise(self, exception_type, graph):
        """Try iterating and topo_sorting graph and expect an exception."""
        self.assertRaises(exception_type, topo_sort, graph)
        self.assertRaises(exception_type, list, TopoSorter(graph).iter_topo_order())

    def assertSortAndIterateOrder(self, graph):
        """Check topo_sort and iter_topo_order is genuinely topological order.

        For every child in the graph, check if it comes after all of it's
        parents.
        """
        sort_result = topo_sort(graph)
        iter_result = list(TopoSorter(graph).iter_topo_order())
        for node, parents in graph:
            for parent in parents:
                if sort_result.index(node) < sort_result.index(parent):
                    self.fail(
                        "parent {} must come before child {}:\n{}".format(
                            parent, node, sort_result
                        )
                    )
                if iter_result.index(node) < iter_result.index(parent):
                    self.fail(
                        "parent {} must come before child {}:\n{}".format(
                            parent, node, iter_result
                        )
                    )

    def test_tsort_empty(self):
        """TopoSort empty list."""
        self.assertSortAndIterate([], [])

    def test_tsort_easy(self):
        """TopoSort list with one node."""
        self.assertSortAndIterate({0: []}.items(), [0])

    def test_tsort_cycle(self):
        """TopoSort traps graph with cycles."""
        self.assertSortAndIterateRaise(GraphCycleError, {0: [1], 1: [0]}.items())

    def test_tsort_cycle_2(self):
        """TopoSort traps graph with longer cycle."""
        self.assertSortAndIterateRaise(
            GraphCycleError, {0: [1], 1: [2], 2: [0]}.items()
        )

    def test_topo_sort_cycle_with_tail(self):
        """TopoSort traps graph with longer cycle."""
        self.assertSortAndIterateRaise(
            GraphCycleError, {0: [1], 1: [2], 2: [3, 4], 3: [0], 4: []}.items()
        )

    def test_tsort_1(self):
        """TopoSort simple nontrivial graph."""
        self.assertSortAndIterate(
            {0: [3], 1: [4], 2: [1, 4], 3: [], 4: [0, 3]}.items(), [3, 0, 4, 1, 2]
        )

    def test_tsort_partial(self):
        """Topological sort with partial ordering.

        Multiple correct orderings are possible, so test for
        correctness, not for exact match on the resulting list.
        """
        self.assertSortAndIterateOrder(
            [
                (0, []),
                (1, [0]),
                (2, [0]),
                (3, [0]),
                (4, [1, 2, 3]),
                (5, [1, 2]),
                (6, [1, 2]),
                (7, [2, 3]),
                (8, [0, 1, 4, 5, 6]),
            ]
        )

    def test_tsort_unincluded_parent(self):
        """Sort nodes, but don't include some parents in the output."""
        self.assertSortAndIterate([(0, [1]), (1, [2])], [1, 0])


class MergeSortTests(TestCase):
    def assertSortAndIterate(
        self, graph, branch_tip, result_list, generate_revno, mainline_revisions=None
    ):
        """Check that merge based sort and iter_topo_order on graph works."""
        value = merge_sort(
            graph,
            branch_tip,
            mainline_revisions=mainline_revisions,
            generate_revno=generate_revno,
        )
        if result_list != value:
            self.assertEqualDiff(pprint.pformat(result_list), pprint.pformat(value))
        self.assertEqual(
            result_list,
            list(
                MergeSorter(
                    graph,
                    branch_tip,
                    mainline_revisions=mainline_revisions,
                    generate_revno=generate_revno,
                ).iter_topo_order()
            ),
        )

    def test_merge_sort_empty(self):
        # sorting of an emptygraph does not error
        self.assertSortAndIterate({}, None, [], False)
        self.assertSortAndIterate({}, None, [], True)
        self.assertSortAndIterate({}, NULL_REVISION, [], False)
        self.assertSortAndIterate({}, NULL_REVISION, [], True)

    def test_merge_sort_not_empty_no_tip(self):
        # merge sorting of a branch starting with None should result
        # in an empty list: no revisions are dragged in.
        self.assertSortAndIterate({0: []}.items(), None, [], False)
        self.assertSortAndIterate({0: []}.items(), None, [], True)

    def test_merge_sort_one_revision(self):
        # sorting with one revision as the tip returns the correct fields:
        # sequence - 0, revision id, merge depth - 0, end_of_merge
        self.assertSortAndIterate({"id": []}.items(), "id", [(0, "id", 0, True)], False)
        self.assertSortAndIterate(
            {"id": []}.items(), "id", [(0, "id", 0, (1,), True)], True
        )

    def test_sequence_numbers_increase_no_merges(self):
        # emit a few revisions with no merges to check the sequence
        # numbering works in trivial cases
        self.assertSortAndIterate(
            {"A": [], "B": ["A"], "C": ["B"]}.items(),
            "C",
            [
                (0, "C", 0, False),
                (1, "B", 0, False),
                (2, "A", 0, True),
            ],
            False,
        )
        self.assertSortAndIterate(
            {"A": [], "B": ["A"], "C": ["B"]}.items(),
            "C",
            [
                (0, "C", 0, (3,), False),
                (1, "B", 0, (2,), False),
                (2, "A", 0, (1,), True),
            ],
            True,
        )

    def test_sequence_numbers_increase_with_merges(self):
        # test that sequence numbers increase across merges
        self.assertSortAndIterate(
            {"A": [], "B": ["A"], "C": ["A", "B"]}.items(),
            "C",
            [
                (0, "C", 0, False),
                (1, "B", 1, True),
                (2, "A", 0, True),
            ],
            False,
        )
        self.assertSortAndIterate(
            {"A": [], "B": ["A"], "C": ["A", "B"]}.items(),
            "C",
            [
                (0, "C", 0, (2,), False),
                (1, "B", 1, (1, 1, 1), True),
                (2, "A", 0, (1,), True),
            ],
            True,
        )

    def test_merge_sort_race(self):
        # A
        # |
        # B-.
        # |\ \
        # | | C
        # | |/
        # | D
        # |/
        # F
        graph = {
            "A": [],
            "B": ["A"],
            "C": ["B"],
            "D": ["B", "C"],
            "F": ["B", "D"],
        }
        self.assertSortAndIterate(
            graph,
            "F",
            [
                (0, "F", 0, (3,), False),
                (1, "D", 1, (2, 2, 1), False),
                (2, "C", 2, (2, 1, 1), True),
                (3, "B", 0, (2,), False),
                (4, "A", 0, (1,), True),
            ],
            True,
        )
        # A
        # |
        # B-.
        # |\ \
        # | X C
        # | |/
        # | D
        # |/
        # F
        graph = {
            "A": [],
            "B": ["A"],
            "C": ["B"],
            "X": ["B"],
            "D": ["X", "C"],
            "F": ["B", "D"],
        }
        self.assertSortAndIterate(
            graph,
            "F",
            [
                (0, "F", 0, (3,), False),
                (1, "D", 1, (2, 1, 2), False),
                (2, "C", 2, (2, 2, 1), True),
                (3, "X", 1, (2, 1, 1), True),
                (4, "B", 0, (2,), False),
                (5, "A", 0, (1,), True),
            ],
            True,
        )

    def test_merge_depth_with_nested_merges(self):
        # the merge depth marker should reflect the depth of the revision
        # in terms of merges out from the mainline
        # revid, depth, parents:
        #  A 0   [D, B]
        #  B  1  [C, F]
        #  C  1  [H]
        #  D 0   [H, E]
        #  E  1  [G, F]
        #  F   2 [G]
        #  G  1  [H]
        #  H 0
        self.assertSortAndIterate(
            {
                "A": ["D", "B"],
                "B": ["C", "F"],
                "C": ["H"],
                "D": ["H", "E"],
                "E": ["G", "F"],
                "F": ["G"],
                "G": ["H"],
                "H": [],
            }.items(),
            "A",
            [
                (0, "A", 0, False),
                (1, "B", 1, False),
                (2, "C", 1, True),
                (3, "D", 0, False),
                (4, "E", 1, False),
                (5, "F", 2, True),
                (6, "G", 1, True),
                (7, "H", 0, True),
            ],
            False,
        )
        self.assertSortAndIterate(
            {
                "A": ["D", "B"],
                "B": ["C", "F"],
                "C": ["H"],
                "D": ["H", "E"],
                "E": ["G", "F"],
                "F": ["G"],
                "G": ["H"],
                "H": [],
            }.items(),
            "A",
            [
                (0, "A", 0, (3,), False),
                (1, "B", 1, (1, 3, 2), False),
                (2, "C", 1, (1, 3, 1), True),
                (3, "D", 0, (2,), False),
                (4, "E", 1, (1, 1, 2), False),
                (5, "F", 2, (1, 2, 1), True),
                (6, "G", 1, (1, 1, 1), True),
                (7, "H", 0, (1,), True),
            ],
            True,
        )

    def test_dotted_revnos_with_simple_merges(self):
        # A         1
        # |\
        # B C       2, 1.1.1
        # | |\
        # D E F     3, 1.1.2, 1.2.1
        # |/ /|
        # G H I     4, 1.2.2, 1.3.1
        # |/ /
        # J K       5, 1.3.2
        # |/
        # L         6
        self.assertSortAndIterate(
            {
                "A": [],
                "B": ["A"],
                "C": ["A"],
                "D": ["B"],
                "E": ["C"],
                "F": ["C"],
                "G": ["D", "E"],
                "H": ["F"],
                "I": ["F"],
                "J": ["G", "H"],
                "K": ["I"],
                "L": ["J", "K"],
            }.items(),
            "L",
            [
                (0, "L", 0, (6,), False),
                (1, "K", 1, (1, 3, 2), False),
                (2, "I", 1, (1, 3, 1), True),
                (3, "J", 0, (5,), False),
                (4, "H", 1, (1, 2, 2), False),
                (5, "F", 1, (1, 2, 1), True),
                (6, "G", 0, (4,), False),
                (7, "E", 1, (1, 1, 2), False),
                (8, "C", 1, (1, 1, 1), True),
                (9, "D", 0, (3,), False),
                (10, "B", 0, (2,), False),
                (11, "A", 0, (1,), True),
            ],
            True,
        )
        # Adding a shortcut from the first revision should not change any of
        # the existing numbers
        self.assertSortAndIterate(
            {
                "A": [],
                "B": ["A"],
                "C": ["A"],
                "D": ["B"],
                "E": ["C"],
                "F": ["C"],
                "G": ["D", "E"],
                "H": ["F"],
                "I": ["F"],
                "J": ["G", "H"],
                "K": ["I"],
                "L": ["J", "K"],
                "M": ["A"],
                "N": ["L", "M"],
            }.items(),
            "N",
            [
                (0, "N", 0, (7,), False),
                (1, "M", 1, (1, 4, 1), True),
                (2, "L", 0, (6,), False),
                (3, "K", 1, (1, 3, 2), False),
                (4, "I", 1, (1, 3, 1), True),
                (5, "J", 0, (5,), False),
                (6, "H", 1, (1, 2, 2), False),
                (7, "F", 1, (1, 2, 1), True),
                (8, "G", 0, (4,), False),
                (9, "E", 1, (1, 1, 2), False),
                (10, "C", 1, (1, 1, 1), True),
                (11, "D", 0, (3,), False),
                (12, "B", 0, (2,), False),
                (13, "A", 0, (1,), True),
            ],
            True,
        )

    def test_end_of_merge_not_last_revision_in_branch(self):
        # within a branch only the last revision gets an
        # end of merge marker.
        self.assertSortAndIterate(
            {
                "A": ["B"],
                "B": [],
            },
            "A",
            [(0, "A", 0, False), (1, "B", 0, True)],
            False,
        )
        self.assertSortAndIterate(
            {
                "A": ["B"],
                "B": [],
            },
            "A",
            [(0, "A", 0, (2,), False), (1, "B", 0, (1,), True)],
            True,
        )

    def test_end_of_merge_multiple_revisions_merged_at_once(self):
        # when multiple branches are merged at once, both of their
        # branch-endpoints should be listed as end-of-merge.
        # Also, the order of the multiple merges should be
        # left-right shown top to bottom.
        # * means end of merge
        # A 0    [H, B, E]
        # B  1   [D, C]
        # C   2  [D]       *
        # D  1   [H]       *
        # E  1   [G, F]
        # F   2  [G]       *
        # G  1   [H]       *
        # H 0    []        *
        self.assertSortAndIterate(
            {
                "A": ["H", "B", "E"],
                "B": ["D", "C"],
                "C": ["D"],
                "D": ["H"],
                "E": ["G", "F"],
                "F": ["G"],
                "G": ["H"],
                "H": [],
            },
            "A",
            [
                (0, "A", 0, False),
                (1, "B", 1, False),
                (2, "C", 2, True),
                (3, "D", 1, True),
                (4, "E", 1, False),
                (5, "F", 2, True),
                (6, "G", 1, True),
                (7, "H", 0, True),
            ],
            False,
        )
        self.assertSortAndIterate(
            {
                "A": ["H", "B", "E"],
                "B": ["D", "C"],
                "C": ["D"],
                "D": ["H"],
                "E": ["G", "F"],
                "F": ["G"],
                "G": ["H"],
                "H": [],
            },
            "A",
            [
                (0, "A", 0, (2,), False),
                (1, "B", 1, (1, 3, 2), False),
                (2, "C", 2, (1, 4, 1), True),
                (3, "D", 1, (1, 3, 1), True),
                (4, "E", 1, (1, 1, 2), False),
                (5, "F", 2, (1, 2, 1), True),
                (6, "G", 1, (1, 1, 1), True),
                (7, "H", 0, (1,), True),
            ],
            True,
        )

    def test_mainline_revs_partial(self):
        # when a mainline_revisions list is passed this must
        # override the graphs idea of mainline, and must also
        # truncate the output to the specified range, if needed.
        # so we test both at once: a mainline_revisions list that
        # disagrees with the graph about which revs are 'mainline'
        # and also truncates the output.
        # graph:
        # A 0 [E, B]
        # B 1 [D, C]
        # C 2 [D]
        # D 1 [E]
        # E 0
        # with a mainline of NONE,E,A (the inferred one) this will show the
        # merge depths above.
        # with a overriden mainline of NONE,E,D,B,A it should show:
        # A 0
        # B 0
        # C 1
        # D 0
        # E 0
        # and thus when truncated to D,B,A it should show
        # A 0
        # B 0
        # C 1
        # because C is brought in by B in this view and D
        # is the terminating revision id
        # this should also preserve revision numbers: C should still be 2.1.1
        self.assertSortAndIterate(
            {"A": ["E", "B"], "B": ["D", "C"], "C": ["D"], "D": ["E"], "E": []},
            "A",
            [
                (0, "A", 0, False),
                (1, "B", 0, False),
                (2, "C", 1, True),
            ],
            False,
            mainline_revisions=["D", "B", "A"],
        )
        self.assertSortAndIterate(
            {"A": ["E", "B"], "B": ["D", "C"], "C": ["D"], "D": ["E"], "E": []},
            "A",
            [
                (0, "A", 0, (4,), False),
                (1, "B", 0, (3,), False),
                (2, "C", 1, (2, 1, 1), True),
            ],
            True,
            mainline_revisions=["D", "B", "A"],
        )

    def test_mainline_revs_with_none(self):
        # a simple test to ensure that a mainline_revs
        # list which goes all the way to None works
        self.assertSortAndIterate(
            {
                "A": [],
            },
            "A",
            [
                (0, "A", 0, True),
            ],
            False,
            mainline_revisions=[None, "A"],
        )
        self.assertSortAndIterate(
            {
                "A": [],
            },
            "A",
            [
                (0, "A", 0, (1,), True),
            ],
            True,
            mainline_revisions=[None, "A"],
        )

    def test_mainline_revs_with_ghost(self):
        # We have a mainline, but the end of it is actually a ghost
        # The graph that is passed to tsort has had ghosts filtered out, but
        # the mainline history has not.
        self.assertSortAndIterate(
            {"B": [], "C": ["B"]}.items(),
            "C",
            [
                (0, "C", 0, (2,), False),
                (1, "B", 0, (1,), True),
            ],
            True,
            mainline_revisions=["A", "B", "C"],
        )

    def test_parallel_root_sequence_numbers_increase_with_merges(self):
        """When there are parallel roots, check their revnos."""
        self.assertSortAndIterate(
            {"A": [], "B": [], "C": ["A", "B"]}.items(),
            "C",
            [
                (0, "C", 0, (2,), False),
                (1, "B", 1, (0, 1, 1), True),
                (2, "A", 0, (1,), True),
            ],
            True,
        )

    def test_revnos_are_globally_assigned(self):
        """Revnos are assigned according to the revision they derive from."""
        # in this test we setup a number of branches that all derive from
        # the first revision, and then merge them one at a time, which
        # should give the revisions as they merge numbers still deriving from
        # the revision were based on.
        # merge 3: J: ['G', 'I']
        # branch 3:
        #  I: ['H']
        #  H: ['A']
        # merge 2: G: ['D', 'F']
        # branch 2:
        #  F: ['E']
        #  E: ['A']
        # merge 1: D: ['A', 'C']
        # branch 1:
        #  C: ['B']
        #  B: ['A']
        # root: A: []
        self.assertSortAndIterate(
            {
                "J": ["G", "I"],
                "I": [
                    "H",
                ],
                "H": ["A"],
                "G": ["D", "F"],
                "F": ["E"],
                "E": ["A"],
                "D": ["A", "C"],
                "C": ["B"],
                "B": ["A"],
                "A": [],
            }.items(),
            "J",
            [
                (0, "J", 0, (4,), False),
                (1, "I", 1, (1, 3, 2), False),
                (2, "H", 1, (1, 3, 1), True),
                (3, "G", 0, (3,), False),
                (4, "F", 1, (1, 2, 2), False),
                (5, "E", 1, (1, 2, 1), True),
                (6, "D", 0, (2,), False),
                (7, "C", 1, (1, 1, 2), False),
                (8, "B", 1, (1, 1, 1), True),
                (9, "A", 0, (1,), True),
            ],
            True,
        )

    def test_roots_and_sub_branches_versus_ghosts(self):
        """Extra roots and their mini branches use the same numbering.

        All of them use the 0-node numbering.
        """
        #       A D   K
        #       | |\  |\
        #       B E F L M
        #       | |/  |/
        #       C G   N
        #       |/    |\
        #       H I   O P
        #       |/    |/
        #       J     Q
        #       |.---'
        #       R
        self.assertSortAndIterate(
            {
                "A": [],
                "B": ["A"],
                "C": ["B"],
                "D": [],
                "E": ["D"],
                "F": ["D"],
                "G": ["E", "F"],
                "H": ["C", "G"],
                "I": [],
                "J": ["H", "I"],
                "K": [],
                "L": ["K"],
                "M": ["K"],
                "N": ["L", "M"],
                "O": ["N"],
                "P": ["N"],
                "Q": ["O", "P"],
                "R": ["J", "Q"],
            }.items(),
            "R",
            [
                (0, "R", 0, (6,), False),
                (1, "Q", 1, (0, 4, 5), False),
                (2, "P", 2, (0, 6, 1), True),
                (3, "O", 1, (0, 4, 4), False),
                (4, "N", 1, (0, 4, 3), False),
                (5, "M", 2, (0, 5, 1), True),
                (6, "L", 1, (0, 4, 2), False),
                (7, "K", 1, (0, 4, 1), True),
                (8, "J", 0, (5,), False),
                (9, "I", 1, (0, 3, 1), True),
                (10, "H", 0, (4,), False),
                (11, "G", 1, (0, 1, 3), False),
                (12, "F", 2, (0, 2, 1), True),
                (13, "E", 1, (0, 1, 2), False),
                (14, "D", 1, (0, 1, 1), True),
                (15, "C", 0, (3,), False),
                (16, "B", 0, (2,), False),
                (17, "A", 0, (1,), True),
            ],
            True,
        )



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import sys


class FG:
    """Unix terminal foreground color codes (16-color)."""

    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    # Bold Foreground
    BOLD_RED = "\033[1;31m"
    BOLD_GREEN = "\033[1;32m"
    BOLD_YELLOW = "\033[1;33m"
    BOLD_BLUE = "\033[1;34m"
    BOLD_MAGENTA = "\033[1;35m"
    BOLD_CYAN = "\033[1;36m"
    BOLD_WHITE = "\033[1;37m"

    NONE = "\033[0m"


class BG:
    """Unix terminal background color codes (16-color)."""

    BLACK = "\033[40m"
    RED = "\033[41m"
    GREEN = "\033[42m"
    YELLOW = "\033[43m"
    BLUE = "\033[44m"
    MAGENTA = "\033[45m"
    CYAN = "\033[46m"
    WHITE = "\033[47m"

    NONE = "\033[0m"


def color_string(s, fg, bg=""):
    return fg + bg + s + FG.NONE


def re_color_string(compiled_pattern, s, fg):
    return compiled_pattern.sub(fg + r"\1" + FG.NONE, s)


def allow_color():
    if os.name != "posix":
        return False
    if not sys.stdout.isatty():
        return False
    try:
        import curses

        curses.setupterm()
        return curses.tigetnum("colors") > 2
    except curses.error:
        return False
